# ***************************************************************************
# *   Copyright (c) 2004 Juergen Riegel <juergen.riegel@web.de>             *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************/

import math
import os
import sys
import tempfile
import unittest
import FreeCAD
from FreeCAD import Base


class ConsoleTestCase(unittest.TestCase):
    def setUp(self):
        self.count = 0

    def testPrint(self):
        FreeCAD.Console.PrintMessage("   Printing message\n")
        FreeCAD.Console.PrintError("   Printing error\n")
        FreeCAD.Console.PrintWarning("   Printing warning\n")
        FreeCAD.Console.PrintLog("   Printing Log\n")

    def testSynchronPrintFromThread(self):
        # http://python-kurs.eu/threads.php
        try:
            import _thread as thread, time
        except Exception:
            import thread, time

        def adder():
            lock.acquire()
            self.count = self.count + 1
            # call of Console method is thread-safe
            FreeCAD.Console.PrintMessage("Call from Python thread: count=" + str(self.count) + "\n")
            lock.release()

        lock = thread.allocate_lock()
        for i in range(10):
            thread.start_new(adder, ())

        time.sleep(3)
        self.assertEqual(self.count, 10, "Synchronization of threads failed")
        FreeCAD.Console.PrintMessage(str(self.count) + "\n")

    def testAsynchronPrintFromThread(self):
        # http://python-kurs.eu/threads.php
        try:
            import _thread as thread, time
        except Exception:
            import thread, time

        def adder():
            self.count = self.count + 1
            # call of Console method is thread-safe
            FreeCAD.Console.PrintMessage(
                "Call from Python thread (not synchronized): count=" + str(self.count) + "\n"
            )

        lock = thread.allocate_lock()
        for i in range(10):
            thread.start_new(adder, ())

        time.sleep(3)
        FreeCAD.Console.PrintMessage(str(self.count) + "\n")

    def tearDown(self):
        pass

    def testILoggerBlocker(self):
        if FreeCAD.GuiUp:
            import QtUnitGui

            QtUnitGui.testILoggerBlocker()


class ParameterTestCase(unittest.TestCase):
    def setUp(self):
        self.TestPar = FreeCAD.ParamGet("System parameter:Test")

    def testGroup(self):
        # FreeCAD.Console.PrintLog("Base::ParameterTestCase::testGroup\n")
        # check on Group creation
        Temp = self.TestPar.GetGroup("44")
        self.assertTrue(self.TestPar.HasGroup("44"), "Test on created group failed")
        # check on Deletion
        self.TestPar.RemGroup("44")
        self.assertTrue(self.TestPar.HasGroup("44"), "A referenced group must not be deleted")
        Temp = 0

    def testGroupNames(self):
        with self.assertRaises(ValueError):
            # no empty groups allowed
            self.TestPar.GetGroup("")
        grp1 = self.TestPar.GetGroup("////Sub1/////Sub2/////")
        grp2 = self.TestPar.GetGroup("Sub1/Sub2")
        self.assertEqual(grp1.GetGroupName(), "Sub2")
        self.assertEqual(grp2.GetGroupName(), "Sub2")

    # check on special conditions
    def testInt(self):
        # check on Int
        self.TestPar.SetInt("44", 4711)
        self.assertEqual(self.TestPar.GetInt("44"), 4711, "In and out error at Int")
        # check on Deletion
        self.TestPar.RemInt("44")
        self.assertEqual(self.TestPar.GetInt("44", 1), 1, "Deletion error at Int")

    def testBool(self):
        # check on Int
        self.TestPar.SetBool("44", 1)
        self.assertEqual(self.TestPar.GetBool("44"), 1, "In and out error at Bool")
        # check on Deletion
        self.TestPar.RemBool("44")
        self.assertEqual(self.TestPar.GetBool("44", 0), 0, "Deletion error at Bool")

    def testFloat(self):
        # check on Int
        self.TestPar.SetFloat("44", 4711.4711)
        self.assertEqual(self.TestPar.GetFloat("44"), 4711.4711, "In and out error at Float")
        # check on Deletion
        self.TestPar.RemFloat("44")
        self.assertEqual(self.TestPar.GetFloat("44", 1.1), 1.1, "Deletion error at Float")

    def testString(self):
        # check on Int
        self.TestPar.SetString("44", "abcdefgh")
        self.assertEqual(self.TestPar.GetString("44"), "abcdefgh", "In and out error at String")
        # check on Deletion
        self.TestPar.RemString("44")
        self.assertEqual(self.TestPar.GetString("44", "hallo"), "hallo", "Deletion error at String")

    def testNesting(self):
        for i in range(50):
            self.TestPar.SetFloat(str(i), 4711.4711)
            self.TestPar.SetInt(str(i), 4711)
            self.TestPar.SetBool(str(i), 1)
            Temp = self.TestPar.GetGroup(str(i))
            for l in range(50):
                Temp.SetFloat(str(l), 4711.4711)
                Temp.SetInt(str(l), 4711)
                Temp.SetBool(str(l), 1)
        Temp = 0

    def testExportImport(self):
        self.TestPar.SetFloat("ExTest", 4711.4711)
        self.TestPar.SetInt("ExTest", 4711)
        self.TestPar.SetString("ExTest", "4711")
        self.TestPar.SetBool("ExTest", 1)
        Temp = self.TestPar.GetGroup("ExTest")
        Temp.SetFloat("ExTest", 4711.4711)
        Temp.SetInt("ExTest", 4711)
        Temp.SetString("ExTest", "4711")
        Temp.SetBool("ExTest", 1)
        TempPath = tempfile.gettempdir() + os.sep + "ExportTest.FCExport"

        self.TestPar.Export(TempPath)
        Temp = self.TestPar.GetGroup("ImportTest")
        Temp.Import(TempPath)
        self.assertEqual(Temp.GetFloat("ExTest"), 4711.4711, "ExportImport error")
        Temp = 0

    def tearDown(self):
        # remove all
        TestPar = FreeCAD.ParamGet("System parameter:Test")
        TestPar.Clear()


class AlgebraTestCase(unittest.TestCase):
    def setUp(self):
        pass

    def testAngle(self):
        v1 = FreeCAD.Vector(0, 0, 0.000001)
        v2 = FreeCAD.Vector(0, 0.000001, 0)
        self.assertAlmostEqual(v1.getAngle(v2), math.pi / 2)
        self.assertAlmostEqual(v2.getAngle(v1), math.pi / 2)

    def testVector2d(self):
        v = FreeCAD.Base.Vector2d(1.0, 1.0)
        v.rotate(math.pi / 2)
        self.assertAlmostEqual(v.x, -1.0)
        self.assertAlmostEqual(v.y, 1.0)

    def testAngleWithNullVector(self):
        v1 = FreeCAD.Vector(0, 0, 0)
        v2 = FreeCAD.Vector(0, 1, 0)
        self.assertTrue(math.isnan(v1.getAngle(v2)))
        self.assertTrue(math.isnan(v2.getAngle(v1)))

    def testMatrix(self):
        m = FreeCAD.Matrix(4, 2, 1, 0, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1)
        u = m.multiply(m.inverse())
        self.assertEqual(u, FreeCAD.Matrix(), "Invalid inverse of matrix")

    def testRotAndMoveMatrix(self):
        m1 = FreeCAD.Matrix()
        m1.move(10, 5, -3)
        m1.rotateY(0.2)
        m2 = FreeCAD.Matrix()
        m2.rotateY(0.2)
        m2.move(10, 5, -3)
        m3 = FreeCAD.Matrix()
        m3.move(10, 5, -3)
        m4 = FreeCAD.Matrix()
        m4.rotateY(0.2)
        self.assertNotEqual(m1, m3 * m4, "Wrong multiplication order")
        self.assertEqual(m1, m4 * m3, "Wrong multiplication order")
        self.assertEqual(m2, m3 * m4, "Wrong multiplication order")
        self.assertNotEqual(m2, m4 * m3, "Wrong multiplication order")

    def testRotationFromMatrix(self):
        rot = FreeCAD.Rotation(45, 30, 0)
        m_r = rot.toMatrix()
        m_r.move(5, 6, 7)
        m_s = FreeCAD.Matrix()
        m_s.scale(1, 1, 3)
        m_s.move(7, 3, 2)
        target_rot = FreeCAD.Rotation(45, 60, 0)
        err = "Non uniform scale has wrong affect non orthogonal rotation"
        self.assertTrue(FreeCAD.Rotation(m_s * m_r).isSame(target_rot, 1e-12), err)
        err = "Right multiplication with non uniform scale must not affect rotation"
        self.assertTrue(FreeCAD.Rotation(m_r * m_s).isSame(rot, 1e-12), err)
        m_r.scale(-2)
        err = "Uniform scale must not affect rotation"
        self.assertTrue(FreeCAD.Rotation(m_r).isSame(rot, 1e-12), err)

    def testRotation(self):
        r = FreeCAD.Rotation(1, 0, 0, 0)  # 180 deg around (1,0,0)
        self.assertEqual(r.Axis, FreeCAD.Vector(1, 0, 0))
        self.assertAlmostEqual(math.fabs(r.Angle), math.fabs(math.pi))

        r = r.multiply(r)  # identity
        self.assertEqual(r.Axis, FreeCAD.Vector(0, 0, 1))
        self.assertAlmostEqual(r.Angle, 0)

        r = FreeCAD.Rotation(1, 0, 0, 0)
        r.Q = (0, 0, 0, 1)  # update axis and angle
        s = FreeCAD.Rotation(0, 0, 0, 1)
        self.assertEqual(r.Axis, s.Axis)
        self.assertAlmostEqual(r.Angle, s.Angle)
        self.assertTrue(r.isSame(s))

        r = FreeCAD.Rotation(1, 0, 0, 0)
        r.Matrix = FreeCAD.Matrix()  # update axis and angle
        s = FreeCAD.Rotation(0, 0, 0, 1)
        self.assertEqual(r.Axis, s.Axis)
        self.assertAlmostEqual(r.Angle, s.Angle)
        self.assertTrue(r.isSame(s))

        r = FreeCAD.Rotation(1, 0, 0, 0)
        r.Axes = (FreeCAD.Vector(0, 0, 1), FreeCAD.Vector(0, 0, 1))  # update axis and angle
        s = FreeCAD.Rotation(0, 0, 0, 1)
        self.assertEqual(r.Axis, s.Axis)
        self.assertAlmostEqual(r.Angle, s.Angle)
        self.assertTrue(r.isSame(s))

        # add 360 deg to angle
        r = FreeCAD.Rotation(FreeCAD.Vector(1, 0, 0), 270)
        s = FreeCAD.Rotation(FreeCAD.Vector(1, 0, 0), 270 + 360)
        self.assertEqual(r.Axis, s.Axis)
        # self.assertAlmostEqual(r.Angle, s.Angle + 2*math.pi)
        self.assertTrue(r.isSame(s))

        # subtract 360 deg from angle using Euler angles
        r = FreeCAD.Rotation(0, 0, 180)
        r.invert()
        s = FreeCAD.Rotation(0, 0, -180)
        self.assertTrue(r.isSame(s))

        # subtract 360 deg from angle using quaternion
        r = FreeCAD.Rotation(1, 0, 0, 0)
        s = FreeCAD.Rotation(-1, 0, 0, 0)
        # angles have the same sign
        if r.Angle * s.Angle > 0:
            self.assertEqual(r.Axis, s.Axis * (-1))
        else:
            self.assertAlmostEqual(r.Angle, -s.Angle)
        self.assertTrue(r.isSame(s))
        r.invert()
        self.assertTrue(r.isSame(s))

        # gimbal lock (north pole)
        r = FreeCAD.Rotation()
        r.setYawPitchRoll(20, 90, 10)
        a = r.getYawPitchRoll()
        s = FreeCAD.Rotation()
        s.setYawPitchRoll(*a)
        self.assertAlmostEqual(a[0], 0.0)
        self.assertAlmostEqual(a[1], 90.0)
        self.assertAlmostEqual(a[2], -10.0)
        self.assertTrue(r.isSame(s, 1e-12))

        # gimbal lock (south pole)
        r = FreeCAD.Rotation()
        r.setYawPitchRoll(20, -90, 10)
        a = r.getYawPitchRoll()
        s = FreeCAD.Rotation()
        s.setYawPitchRoll(*a)
        self.assertAlmostEqual(a[0], 0.0)
        self.assertAlmostEqual(a[1], -90.0)
        self.assertAlmostEqual(a[2], 30.0)
        self.assertTrue(r.isSame(s, 1e-12))

    def testInverted(self):
        p = FreeCAD.Placement()
        p.Rotation.Angle = math.pi / 2
        self.assertEqual(abs(p.inverse().Rotation.Angle), p.Rotation.Angle)

    def testDecomposeScale(self):
        mat = FreeCAD.Matrix()
        mat.scale(FreeCAD.Vector(2, 3, 4))
        res = mat.decompose()
        self.assertTrue(res[0].isUnity())
        self.assertEqual(res[1], mat)
        self.assertTrue(res[2].isUnity())
        self.assertTrue(res[3].isUnity())

    def testDecomposeRotation(self):
        mat = FreeCAD.Matrix()
        mat.rotateX(1)
        mat.rotateY(1)
        mat.rotateZ(1)
        res = mat.decompose()
        self.assertTrue(res[0].isUnity())
        self.assertTrue(res[1].isUnity(1e-12))
        self.assertEqual(res[2], mat)
        self.assertTrue(res[3].isUnity())

    def testDecomposeMove(self):
        mat = FreeCAD.Matrix()
        mat.move(FreeCAD.Vector(1, 2, 3))
        res = mat.decompose()
        self.assertTrue(res[0].isUnity())
        self.assertTrue(res[1].isUnity())
        self.assertTrue(res[2].isUnity())
        self.assertEqual(res[3], mat)

    def testMatrixDecompose(self):
        mat = FreeCAD.Matrix()
        mat.A21 = 1.0
        mat.A14 = 1.0
        mat.A24 = 2.0
        mat.A34 = 3.0
        res = mat.decompose()

        mul = res[3] * res[2] * res[1] * res[0]
        self.assertEqual(mul, mat)

        # shearing
        self.assertAlmostEqual(res[0].determinant(), 1.0)

        # scaling
        self.assertEqual(res[1].hasScale(), FreeCAD.ScaleType.NonUniformLeft)

        # rotation
        rot1 = FreeCAD.Rotation(res[2])
        rot2 = FreeCAD.Rotation(FreeCAD.Vector(0, 0, 1), 45)
        self.assertTrue(rot1.isSame(rot2, 1e-12))

        # translation
        plm = FreeCAD.Placement(res[3])
        self.assertEqual(plm.Base, FreeCAD.Vector(1, 2, 3))

    def testYawPitchRoll(self):
        def getYPR1(yaw, pitch, roll):
            r = FreeCAD.Rotation()
            r.setYawPitchRoll(yaw, pitch, roll)
            return r

        def getYPR2(yaw, pitch, roll):
            rx = FreeCAD.Rotation()
            ry = FreeCAD.Rotation()
            rz = FreeCAD.Rotation()

            rx.Axis = FreeCAD.Vector(1, 0, 0)
            ry.Axis = FreeCAD.Vector(0, 1, 0)
            rz.Axis = FreeCAD.Vector(0, 0, 1)

            rx.Angle = math.radians(roll)
            ry.Angle = math.radians(pitch)
            rz.Angle = math.radians(yaw)

            return rz.multiply(ry).multiply(rx)

        angles = []
        angles.append((10, 10, 10))
        angles.append((13, 45, -24))
        angles.append((10, -90, 20))

        for i in angles:
            r = getYPR1(*i)
            s = getYPR2(*i)
            self.assertTrue(r.isSame(s, 1e-12))

    def testBounding(self):
        b = FreeCAD.BoundBox()
        b.setVoid()
        self.assertFalse(b.isValid(), "Bbox is not invalid")
        b.add(0, 0, 0)
        self.assertTrue(b.isValid(), "Bbox is invalid")
        self.assertEqual(b.XLength, 0, "X length > 0")
        self.assertEqual(b.YLength, 0, "Y length > 0")
        self.assertEqual(b.ZLength, 0, "Z length > 0")
        self.assertEqual(b.Center, FreeCAD.Vector(0, 0, 0), "Center is not at (0,0,0)")
        self.assertTrue(b.isInside(b.Center), "Center is not inside Bbox")
        b.add(2, 2, 2)
        self.assertTrue(
            b.isInside(b.getIntersectionPoint(b.Center, FreeCAD.Vector(0, 1, 0))),
            "Intersection point is not inside Bbox",
        )
        self.assertTrue(b.intersect(b), "Bbox doesn't intersect with itself")
        self.assertFalse(
            b.intersected(FreeCAD.BoundBox(4, 4, 4, 6, 6, 6)).isValid(),
            "Bbox should not intersect with Bbox outside",
        )
        self.assertEqual(
            b.intersected(FreeCAD.BoundBox(-2, -2, -2, 2, 2, 2)).Center,
            b.Center,
            "Bbox is not a full subset",
        )

    def testMultLeftOrRight(self):
        doc = FreeCAD.newDocument()
        obj = doc.addObject("App::FeatureTestPlacement")

        p1 = Base.Placement()
        p1.Base = Base.Vector(10, 10, 10)

        p2 = Base.Placement()
        p2.Rotation.Angle = math.radians(90)

        obj.Input1 = p1
        obj.Input2 = p2
        doc.recompute()

        self.assertTrue(obj.MultRight.isSame(p1 * p2))
        self.assertFalse(obj.MultRight.isSame(p2 * p1))
        self.assertTrue(obj.MultLeft.isSame(p2 * p1))
        self.assertFalse(obj.MultLeft.isSame(p1 * p2))

        FreeCAD.closeDocument(doc.Name)


class MatrixTestCase(unittest.TestCase):
    def setUp(self):
        self.mat = FreeCAD.Matrix()

    def testOrder(self):
        self.mat = FreeCAD.Matrix(1.0, 2.0, 3.0, 4.0)
        self.assertEqual(self.mat.A11, 1.0)
        self.assertEqual(self.mat.A12, 2.0)
        self.assertEqual(self.mat.A13, 3.0)
        self.assertEqual(self.mat.A14, 4.0)

    def testScalar(self):
        res = self.mat * 0.0
        for i in range(16):
            self.assertEqual(res.A[i], 0.0)

    def testAddition(self):
        res1 = self.mat * 2.0
        res2 = self.mat + self.mat
        for i in range(16):
            self.assertEqual(res1.A[i], res2.A[i])

    def testMinus(self):
        res = self.mat - self.mat
        for i in range(16):
            self.assertEqual(res.A[i], 0.0)

    def testVector(self):
        vec = FreeCAD.Vector(1, 1, 1)
        vec = self.mat * vec
        self.assertEqual(vec.x, 1.0)
        self.assertEqual(vec.y, 1.0)
        self.assertEqual(vec.z, 1.0)

    def testVectorMult(self):
        vec = FreeCAD.Vector(1, 1, 1)
        with self.assertRaises(TypeError):
            vec * "string"

    def testRotation(self):
        rot = FreeCAD.Rotation()
        res = self.mat * rot
        self.assertEqual(type(res), FreeCAD.Matrix)

    def testPlacement(self):
        plm = FreeCAD.Placement()
        res = self.mat * plm
        self.assertEqual(type(res), FreeCAD.Matrix)

    def testMatrix(self):
        mat = FreeCAD.Matrix()
        res = self.mat * mat
        self.assertEqual(type(res), FreeCAD.Matrix)

    def testMatrixPlacementMatrix(self):
        # Example taken from https://forum.freecad.org/viewtopic.php?f=3&t=61000
        mat = FreeCAD.Matrix(
            -0.470847778020266,
            0.8150598976807029,
            0.3376088628746235,
            -11.25290913640202,
            -0.8822144756796808,
            -0.4350066260577338,
            -0.180185641360483,
            -2876.45492562325,
            1.955470978815492e-9,
            -0.3826834326750831,
            0.923879538425552,
            941.3822018176414,
        )
        plm = FreeCAD.Placement(mat)
        mat = plm.toMatrix()
        self.assertEqual(mat.hasScale(), FreeCAD.ScaleType.NoScaling)

    def testAnything(self):
        with self.assertRaises(NotImplementedError):
            self.mat * "string"

    def testUnity(self):
        mat = FreeCAD.Matrix(2, 0, 0, 0, 0, 1, 0, 0, 0, 0, 2, 0, 0, 0, 0, -1)
        self.assertFalse(mat.isUnity())
        mat.unity()
        self.assertTrue(mat.isUnity())

    def testPower(self):
        mat = FreeCAD.Matrix(2, 0, 0, 0, 0, 1, 0, 0, 0, 0, 2, 0, 0, 0, 0, -1)
        with self.assertRaises(NotImplementedError):
            mat ** "string"

        mat2 = mat**0
        self.assertTrue(mat2.isUnity())
        self.assertEqual(mat**-1, mat.inverse())
        self.assertEqual(mat**1, mat)
        self.assertEqual(mat**2, mat * mat)
        self.assertEqual(mat**3, mat * mat * mat)
        mat.nullify()
        with self.assertRaises(RuntimeError):
            mat**-1

    def testScale(self):
        self.mat.scale(1.0, 2.0, 3.0)
        self.assertEqual(self.mat.determinant(), 6.0)
        self.assertEqual(self.mat.hasScale(), FreeCAD.ScaleType.NonUniformLeft)
        self.mat.unity()
        self.assertEqual(self.mat.hasScale(), FreeCAD.ScaleType.NoScaling)
        self.mat.scale(2.0, 2.0, 2.0)
        self.assertEqual(self.mat.hasScale(), FreeCAD.ScaleType.Uniform)
        self.mat.rotateX(1.0)
        self.assertEqual(self.mat.hasScale(), FreeCAD.ScaleType.Uniform)
        self.mat.scale(1.0, 2.0, 3.0)
        self.assertEqual(self.mat.hasScale(), FreeCAD.ScaleType.NonUniformLeft)
        self.mat.unity()
        self.mat.scale(1.0, 2.0, 3.0)
        self.mat.rotateX(1.0)
        self.assertEqual(self.mat.hasScale(), FreeCAD.ScaleType.NonUniformRight)
        self.mat.unity()
        self.mat.setCol(0, FreeCAD.Vector(1, 2, 3))
        self.mat.setCol(1, FreeCAD.Vector(1, 2, 3))
        self.mat.setCol(2, FreeCAD.Vector(1, 2, 3))
        self.assertEqual(self.mat.hasScale(), FreeCAD.ScaleType.Other)
        self.mat.unity()
        self.mat.setRow(0, FreeCAD.Vector(1, 2, 3))
        self.mat.setRow(1, FreeCAD.Vector(1, 2, 3))
        self.mat.setRow(2, FreeCAD.Vector(1, 2, 3))
        self.assertEqual(self.mat.hasScale(), FreeCAD.ScaleType.Other)
        self.mat.unity()
        self.mat.scale(-1.0)
        self.assertEqual(self.mat.hasScale(), FreeCAD.ScaleType.Uniform)
        self.mat.scale(-2.0)
        self.assertEqual(self.mat.hasScale(), FreeCAD.ScaleType.Uniform)

    def testShearing(self):
        self.mat.setRow(1, FreeCAD.Vector(0, 1, 1))
        self.assertEqual(self.mat.hasScale(), FreeCAD.ScaleType.Other)

    def testMultLeftOrRight(self):
        mat1 = FreeCAD.Matrix()
        mat1.rotateX(1.0)

        mat2 = FreeCAD.Matrix()
        mat2.scale(1.0, 2.0, 3.0)
        self.assertEqual((mat1 * mat2).hasScale(), FreeCAD.ScaleType.NonUniformRight)
        self.assertEqual((mat2 * mat1).hasScale(), FreeCAD.ScaleType.NonUniformLeft)

    def testNull(self):
        self.assertFalse(self.mat.isNull())
        self.mat.nullify()
        self.assertTrue(self.mat.isNull())

    def testUnity(self):
        self.assertTrue(self.mat.isUnity())
        self.mat.nullify()
        self.assertFalse(self.mat.isUnity())
        self.mat.unity()
        self.assertTrue(self.mat.isUnity())

    def testColRow(self):
        with self.assertRaises(TypeError):
            self.mat.col("string")
        with self.assertRaises(TypeError):
            self.mat.row("string")
        self.assertEqual(type(self.mat.col(0)), FreeCAD.Vector)
        self.assertEqual(type(self.mat.row(0)), FreeCAD.Vector)
        self.mat.setCol(0, FreeCAD.Vector(1, 0, 0))
        self.mat.setRow(0, FreeCAD.Vector(1, 0, 0))

    def testDiagonal(self):
        self.mat.scale(2.0, 2.0, 2.0)
        self.assertEqual(self.mat.diagonal(), FreeCAD.Vector(2.0, 2.0, 2.0))

    def testNumberProtocol(self):
        with self.assertRaises(NotImplementedError):
            self.mat / 2.0
        with self.assertRaises(NotImplementedError):
            self.mat % 2.0
        with self.assertRaises(NotImplementedError):
            divmod(self.mat, 2.0)
        with self.assertRaises(NotImplementedError):
            float(self.mat)
        with self.assertRaises(NotImplementedError):
            int(self.mat)
        with self.assertRaises(NotImplementedError):
            self.mat | self.mat
        with self.assertRaises(NotImplementedError):
            self.mat & self.mat
        with self.assertRaises(NotImplementedError):
            self.mat ^ self.mat
        with self.assertRaises(NotImplementedError):
            self.mat << 2
        with self.assertRaises(NotImplementedError):
            self.mat >> 2
        with self.assertRaises(NotImplementedError):
            ~self.mat
        with self.assertRaises(NotImplementedError):
            abs(self.mat)
        self.assertEqual(+self.mat, self.mat)
        self.assertEqual(-self.mat, self.mat * -1)
        self.assertTrue(bool(self.mat))


class FileSystem(unittest.TestCase):
    def testEncoding(self):
        self.assertEqual(sys.getfilesystemencoding(), "utf-8")



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2025 Furgo                                              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

import FreeCAD
from bimtests import TestArchBase
import Arch
import ArchWindow  # For ArchWindow._Window proxy class
import Part
import Draft
import Sketcher


class TestArchWindow(TestArchBase.TestArchBase):

    def test_create_no_args(self):
        """Test creating a window with no arguments."""
        window = Arch.makeWindow(name="Window_NoArgs")
        self.assertIsNotNone(window)
        self.assertTrue(
            hasattr(window, "Proxy") and isinstance(window.Proxy, ArchWindow._Window),
            "Window proxy is not of expected ArchWindow._Window type",
        )
        self.assertEqual(window.Label, "Window_NoArgs")
        self.assertEqual(window.IfcType, "Window")
        self.assertIsNone(window.Base)
        self.assertEqual(len(window.WindowParts), 0)
        self.assertTrue(window.Shape.isNull())
        self.document.recompute()
        self.assertTrue(window.Shape.isNull())

    def test_create_from_sketch_single_wire_default_parts(self):
        """Test creating a window from a single-wire sketch, relying on default parts."""
        sketch = self._create_sketch_with_wires("SketchSingle", [(0, 0, 1000, 1200)])
        window = Arch.makeWindow(baseobj=sketch, name="Window_SketchSingle_Default")
        self.document.recompute()

        self.assertEqual(window.Base, sketch)
        self.assertTrue(
            len(window.WindowParts) >= 5, "Should have at least one default part (5 elements)."
        )
        self.assertEqual(window.WindowParts[0], "Default")
        self.assertEqual(
            window.WindowParts[1],
            "Solid panel",
            "Default part type incorrect for single-wire sketch.",
        )
        self.assertIn("Wire0", window.WindowParts[2])
        self.assertFalse(window.Shape.isNull())
        self.assertGreater(len(window.Shape.Solids), 0)

    def test_create_from_sketch_two_wires_default_parts(self):
        """Test creating a window from two-wire sketch (concentric), relying on default parts."""
        sketch = self._create_sketch_with_wires(
            "SketchTwoWires", [(0, 0, 1000, 1200), (100, 100, 800, 1000)]
        )
        window = Arch.makeWindow(baseobj=sketch, name="Window_SketchTwo_Default")
        self.document.recompute()

        self.assertEqual(window.Base, sketch)
        self.assertGreaterEqual(len(window.WindowParts), 5)
        self.assertEqual(window.WindowParts[0], "Default")
        self.assertEqual(
            window.WindowParts[1], "Frame", "Default type for multi-wire should be Frame."
        )
        self.assertIn("Wire0", window.WindowParts[2])
        self.assertIn("Wire1", window.WindowParts[2])
        self.assertFalse(window.Shape.isNull())
        self.assertGreater(len(window.Shape.Solids), 0)

    def test_sketch_named_constraints_driven_by_window_props(self):
        """Test that window Width/Height properties drive sketch's named constraints."""
        sketch_width, sketch_height = 800.0, 1000.0
        sketch = self._create_sketch_with_named_constraints(
            "SketchNamed", sketch_width, sketch_height
        )

        window = Arch.makeWindow(baseobj=sketch, name="Window_NamedSketch")
        # Set window Width/Height to ensure they become the drivers if sketch also has the constraints.
        # They need to be set explicitly after creating the window, as they are not automatically initialized
        # from the sketch.
        window.Width = sketch_width
        window.Height = sketch_height
        self.document.recompute()

        self.assertEqual(sketch.getDatum("Width").Value, sketch_width)
        self.assertEqual(sketch.getDatum("Height").Value, sketch_height)

        new_win_width = 1200.0
        window.Width = new_win_width
        self.document.recompute()
        self.assertEqual(
            sketch.getDatum("Width").Value,
            new_win_width,
            "Window.Width should drive sketch 'Width' constraint.",
        )

        new_win_height = 1500.0
        window.Height = new_win_height
        self.document.recompute()
        self.assertEqual(
            sketch.getDatum("Height").Value,
            new_win_height,
            "Window.Height should drive sketch 'Height' constraint.",
        )

    def test_create_from_sketch_with_custom_parts(self):
        """Test creating a window from sketch with explicit custom parts."""
        sketch = self._create_sketch_with_wires(
            "SketchCustom", [(0, 0, 1200, 1000), (100, 100, 1000, 800)]
        )
        custom_parts = [
            "OuterFrame",
            "Frame",
            "Wire0,Wire1",
            "70",
            "0",
            "GlassPane",
            "Glass panel",
            "Wire1",
            "20",
            "25",
        ]
        window = Arch.makeWindow(baseobj=sketch, parts=custom_parts, name="Window_Custom")
        self.document.recompute()

        self.assertEqual(window.Base, sketch)
        self.assertEqual(list(window.WindowParts), custom_parts)
        self.assertFalse(window.Shape.isNull())
        self.assertEqual(len(window.Shape.Solids), 2, "Expected two solids for frame and glass.")

    def test_custom_parts_with_plus_v_references(self):
        """Test creating a window with custom parts using '+V' to reference window.Frame and window.Offset."""
        sketch = self._create_sketch_with_wires("SketchPlusV", [(0, 0, 1000, 800)])
        frame_val = 60.0
        offset_val = 10.0

        custom_parts_plus_v = ["MainFrame", "Frame", "Wire0", "50+V", "5+V"]
        window = Arch.makeWindow(baseobj=sketch, parts=custom_parts_plus_v, name="Window_PlusV")
        window.Frame = frame_val
        window.Offset = offset_val
        self.document.recompute()

        self.assertFalse(window.Shape.isNull())
        self.assertGreater(len(window.Shape.Solids), 0)

    def test_create_with_width_height_no_baseobj_initially(self):
        """
        Test Arch.makeWindow(width, height) current behavior regarding window.Base.
        It verifies that window.Base is not automatically created and remains None,
        and that Width/Height properties are correctly set on the window object.
        """
        win_w, win_h = 1000.0, 1200.0
        window_name = "Window_W_H_NoAutoBase"
        window = Arch.makeWindow(width=win_w, height=win_h, name=window_name)

        # 1. Check initial state after makeWindow call
        self.assertEqual(window.Label, window_name)
        self.assertIsNone(
            window.Base, "Immediately after makeWindow(W,H), window.Base should be None."
        )
        self.assertTrue(
            window.Shape.isNull(), "Initially, window.Shape should be null as no Base or Parts yet."
        )
        self.assertAlmostEqual(
            window.Width.Value,
            win_w,
            places=5,
            msg="Window.Width property not correctly set by makeWindow.",
        )
        self.assertAlmostEqual(
            window.Height.Value,
            win_h,
            places=5,
            msg="Window.Height property not correctly set by makeWindow.",
        )

        # 2. Perform a document recompute
        # This should trigger window.execute() -> ArchComponent.ensureBase()
        self.document.recompute()

        # 3. Assert the CURRENT BEHAVIOR: window.Base remains None
        # The original ArchComponent.ensureBase does not create a sketch if Base is None
        # and only Width/Height are provided to the object.
        self.assertIsNone(
            window.Base,
            "Current Behavior: window.Base should remain None even after recomputes for makeWindow(W,H).",
        )

        # 4. Consequently, window.Shape should still be null.
        self.assertTrue(
            window.Shape.isNull(), "window.Shape should remain null if window.Base was not created."
        )

        # 5. Attempting to set parts that rely on a Base sketch (e.g., "Wire0")
        #    should not produce a shape if Base is None.
        window.WindowParts = ["Default", "Frame", "Wire0", "50", "0"]
        self.document.recompute()
        self.assertTrue(
            window.Shape.isNull(),
            "window.Shape should still be null if WindowParts reference Wire0 from a non-existent Base.",
        )

    def test_window_in_wall_creates_opening(self):
        """
        Test if a window hosted in a wall creates a geometric opening,
        verifying changes in Volume and VerticalArea.
        Manually sets win.Width and win.Height after Arch.makeWindow(sk)
        as current Arch.makeWindow(sk) behavior does not initialize these
        properties from the sketch. It expects win.Placement to remain identity.
        """
        # 1. Create Wall
        wall_length = 3000.0
        wall_thickness = 200.0
        wall_height = 2400.0

        line = Draft.makeLine(FreeCAD.Vector(0, 0, 0), FreeCAD.Vector(wall_length, 0, 0))
        self.document.recompute()

        wall = Arch.makeWall(
            line,
            width=wall_thickness,
            height=wall_height,
            align="Left",
            name="TestWall_ForOpening_Args",
        )
        self.document.recompute()

        initial_wall_volume = wall.Shape.Volume
        initial_wall_vertical_area = wall.VerticalArea.Value

        # 2. Create Sketch for Window profile
        sketch_profile_width = 800.0
        sketch_profile_height = 1000.0
        sk = self._create_sketch_with_wires(
            "WindowSketch_Args", [(0, 0, sketch_profile_width, sketch_profile_height)]
        )

        # Position and orient the sketch in 3D space
        sk.Placement.Rotation = FreeCAD.Rotation(FreeCAD.Vector(1, 0, 0), 90)
        win_global_x_start = (wall.Length.Value - sketch_profile_width) / 2
        win_global_z_start = 900.0
        sk.Placement.Base = FreeCAD.Vector(win_global_x_start, 0, win_global_z_start)
        self.document.recompute()

        # 3. Create Window from sketch
        win = Arch.makeWindow(sk, name="WindowInWall_Args")

        # Manually set Width and Height to match the sketch profile dimensions
        win.Width = sketch_profile_width
        win.Height = sketch_profile_height
        # win.Placement remains identity

        win.HoleDepth = 0  # Use "smart" hole depth calculation
        win.WindowParts = ["DefaultFrame", "Frame", "Wire0", "60", "0"]

        win.recompute()
        self.document.recompute()

        self.assertFalse(win.Shape.isNull(), "Window must have a valid shape.")
        self.assertEqual(
            win.Placement,
            FreeCAD.Placement(),
            f"Window.Placement should be identity; got {win.Placement}",
        )
        self.assertAlmostEqual(win.Width.Value, sketch_profile_width, 5)
        self.assertAlmostEqual(win.Height.Value, sketch_profile_height, 5)

        # 4. Add Window to Wall
        Arch.addComponents(win, host=wall)
        self.document.recompute()

        # 5. Assertions
        # Wall.Subtractions PropertyLinkList check (informational, expected to be empty)
        wall_subtractions_prop = wall.Subtractions if hasattr(wall, "Subtractions") else []
        self.assertEqual(
            len(wall_subtractions_prop),
            0,
            f"Wall.Subtractions property list expected to be empty, but: "
            f"{wall_subtractions_prop}",
        )

        # Geometric checks
        current_wall_volume = wall.Shape.Volume
        self.assertLess(
            current_wall_volume,
            initial_wall_volume,
            f"Wall volume should decrease. Before: {initial_wall_volume}, "
            f"After: {current_wall_volume}",
        )

        ideal_removed_volume = win.Width.Value * win.Height.Value * wall.Width.Value
        self.assertAlmostEqual(
            current_wall_volume,
            initial_wall_volume - ideal_removed_volume,
            places=3,
            msg=f"Wall volume cut not as expected. Initial: {initial_wall_volume}, "
            f"Current: {current_wall_volume}, IdealRemoved: {ideal_removed_volume}",
        )

        # Check VerticalArea
        current_wall_vertical_area = wall.VerticalArea.Value

        window_opening_area_on_one_face = win.Width.Value * win.Height.Value
        area_of_side_reveals = 2 * win.Height.Value * wall.Width.Value
        expected_wall_vertical_area_after = (
            initial_wall_vertical_area
            - (2 * window_opening_area_on_one_face)
            + area_of_side_reveals
        )

        self.assertAlmostEqual(
            current_wall_vertical_area,
            expected_wall_vertical_area_after,
            places=3,
            msg=f"Wall VerticalArea change not as expected. Initial: "
            f"{initial_wall_vertical_area}, Current: {current_wall_vertical_area}, "
            f"ExpectedAfter: {expected_wall_vertical_area_after}",
        )

    def test_clone_window(self):
        """Test cloning an Arch.Window object.

        Notes:
        - The clone's name is automatically generated, the `name` argument is ignored.
        - The clone's WindowParts and other properties are always empty, despite the original having them.

        """
        sketch = self._create_sketch_with_wires("OriginalSketch", [(0, 0, 600, 800)])
        original_parts = ["MainFrame", "Frame", "Wire0", "50", "10"]
        original_window = Arch.makeWindow(
            baseobj=sketch, parts=original_parts, name="OriginalWindow"
        )
        original_window.Frame = 60.0
        self.document.recompute()

        self.assertEqual(original_window.Label, "OriginalWindow")

        cloned_window = Arch.makeWindow(baseobj=original_window)
        self.document.recompute()

        self.assertIsNotNone(cloned_window)
        self.assertIsNotNone(cloned_window.CloneOf)
        self.assertEqual(cloned_window.CloneOf, original_window)

        self.assertEqual(cloned_window.IfcType, original_window.IfcType)

        self.assertFalse(cloned_window.Shape.isNull())
        self.assertAlmostEqual(
            cloned_window.Shape.Volume,
            original_window.Shape.Volume,
            delta=1e-5,
            msg="Cloned window volume should match original.",
        )

    def test_create_window_on_xz_plane(self):
        """Test creating a window oriented on the XZ (vertical) plane."""

        # Create a a frame-like profile.
        sketch = self._create_sketch_with_wires(
            "Sketch_XZ_Plane", [(0, 0, 1000, 1200), (100, 100, 800, 1000)]
        )

        # Orient the sketch to the XZ plane.
        sketch.Placement.Rotation = FreeCAD.Rotation(FreeCAD.Vector(1, 0, 0), 90)
        self.document.recompute()

        # Create the window using explicit parts.
        window = Arch.makeWindow(baseobj=sketch, name="Window_XZ_Plane")
        window.WindowParts = [
            "Frame",
            "Frame",
            "Wire0,Wire1",
            "60",
            "0",  # A 60mm thick frame
            "Glass",
            "Glass panel",
            "Wire1",
            "10",
            "25",  # A 10mm thick glass pane, offset by 25mm
        ]
        self.document.recompute()

        # Check the resulting geometry's orientation and dimensions.
        self.assertFalse(window.Shape.isNull(), "Window shape should not be null.")
        self.assertEqual(
            len(window.Shape.Solids), 2, "Window should contain two solids (frame and glass)."
        )

        bb = window.Shape.BoundBox

        # The window's overall "thickness" is determined by the frame (60mm).
        # Its "width" (X) and "height" (Z) should be larger.
        self.assertGreater(
            bb.XLength,
            bb.YLength,
            "Window XLength (width) should be greater than YLength (thickness).",
        )
        self.assertGreater(
            bb.ZLength,
            bb.YLength,
            "Window ZLength (height) should be greater than YLength (thickness).",
        )

        # Verify the overall thickness is correct (60mm, from the Frame component).
        self.assertAlmostEqual(
            bb.YLength, 60.0, places=5, msg="Window thickness (YLength) is incorrect."
        )

    def test_opening_property_rotates_component(self):
        """Test that setting the Opening property rotates a hinged component."""
        # Create a window from a preset with a hinge
        window = Arch.makeWindowPreset(
            "Open 1-pane", width=1000, height=1200, h1=50, h2=50, h3=0, w1=100, w2=50, o1=0, o2=50
        )
        self.document.recompute()

        # The preset creates an outer frame, an inner opening frame, and glass.
        self.assertEqual(len(window.Shape.Solids), 3, "Preset should create three solids.")

        # Solid[1] is the inner, opening frame.
        opening_pane = window.Shape.Solids[1]

        # Get initial position
        initial_center = opening_pane.CenterOfMass

        # Set opening to 50%. The 'Opening' property is App::PropertyPercent, which expects an integer.
        window.Opening = 50

        self.document.recompute()

        # Get new position of the same component
        new_opening_pane = window.Shape.Solids[1]
        new_center = new_opening_pane.CenterOfMass

        # Assert that the Z-coordinate has changed, indicating rotation around a horizontal hinge.
        self.assertNotAlmostEqual(
            initial_center.z,
            new_center.z,
            places=3,
            msg="The Z-coordinate of the opening pane's center should change upon rotation.",
        )
        # The Y-coordinate should remain largely unchanged for a bottom-hinged (awning-style) window.
        self.assertAlmostEqual(initial_center.y, new_center.y, places=3)

    def test_symbol_plan_creates_wire_geometry(self):
        """Test that enabling SymbolPlan adds 2D wire geometry to the window's shape."""
        # Create a window with an opening mode
        window = Arch.makeWindowPreset(
            "Open 1-pane", width=1000, height=1200, h1=50, h2=50, h3=0, w1=100, w2=50, o1=0, o2=50
        )

        # SymbolPlan is True by default for presets with hinges, but we set it explicitly
        window.SymbolPlan = True

        self.document.recompute()

        # Assert that the shape contains both solids and the 2D symbol edges.
        self.assertIsInstance(window.Shape, Part.Compound, "Window shape should be a compound.")

        # Find symbol edges by finding edges that do not belong to any face in the shape
        face_edge_hashes = {e.hashCode() for face in window.Shape.Faces for e in face.Edges}
        all_edge_hashes = {e.hashCode() for e in window.Shape.Edges}
        symbol_edge_hashes = all_edge_hashes - face_edge_hashes

        self.assertEqual(len(window.Shape.Solids), 3, "Window shape should contain 3 solids.")
        self.assertEqual(
            len(symbol_edge_hashes), 2, "Expected 2 edges for the plan symbol (arc and line)."
        )

    def test_custom_subvolume_creates_opening(self):
        """Test that a custom Subvolume shape correctly creates an opening in a host wall."""
        # Create a wall and store its initial state
        wall_base = Draft.makeLine(FreeCAD.Vector(0, 0, 0), FreeCAD.Vector(4000, 0, 0))
        wall = Arch.makeWall(wall_base, width=200, height=3000, align="Left")
        self.document.recompute()
        initial_wall_volume = wall.Shape.Volume
        initial_wall_shape = (
            wall.Shape.copy()
        )  # Store initial shape for accurate intersection calculation

        # Create a simple window
        window_sketch = self._create_sketch_with_wires("WinSketch", [(0, 0, 10, 10)])
        window = Arch.makeWindow(window_sketch)

        # Create a custom cutting box as the window's Subvolume
        cutting_box = self.document.addObject("Part::Box", "CuttingBox")
        cutting_box.Length = 800
        cutting_box.Width = 300  # Wider than wall
        cutting_box.Height = 1000
        cutting_box.Placement.Base = FreeCAD.Vector(1600, -50, 800)
        self.document.recompute()

        window.Subvolume = cutting_box

        # Add window to wall to create opening using Arch.addComponents
        Arch.addComponents(window, host=wall)
        self.document.recompute()

        # Assert volume has decreased by the intersecting part of the box's volume
        expected_volume_change = cutting_box.Shape.common(initial_wall_shape).Volume
        final_wall_volume = wall.Shape.Volume
        self.assertAlmostEqual(
            final_wall_volume,
            initial_wall_volume - expected_volume_change,
            places=3,
            msg="Wall volume did not decrease correctly after applying custom Subvolume.",
        )

    def test_door_preset_sets_ifctype_door(self):
        """Test that creating a window from a 'door' preset correctly sets its IfcType to 'Door'."""
        # Create a door using a preset with non-zero values for all required parameters
        door = Arch.makeWindowPreset(
            "Simple door", width=900, height=2100, h1=50, h2=50, h3=0, w1=100, w2=40, o1=0, o2=0
        )
        self.assertIsNotNone(door, "makeWindowPreset for a door should create an object.")
        self.document.recompute()

        # Assert IfcType is "Door"
        self.assertEqual(door.IfcType, "Door", "IfcType should be 'Door' for a door preset.")

        # Assert component count
        # A simple door preset has an outer frame and a solid door panel
        self.assertEqual(
            len(door.Shape.Solids),
            2,
            "A simple door should have 2 solid components (frame and panel).",
        )

    def test_cloned_window_in_wall_creates_opening(self):
        """Tests if a cloned Arch.Window, when hosted in a wall, creates a geometric opening."""

        # Create the host wall
        wall_length = 3000.0
        wall_thickness = 200.0
        wall_height = 2500.0
        wall_base = Draft.makeLine(FreeCAD.Vector(0, 0, 0), FreeCAD.Vector(wall_length, 0, 0))
        wall = Arch.makeWall(
            wall_base, width=wall_thickness, height=wall_height, name="WallForClonedWindow"
        )
        self.document.recompute()
        initial_wall_volume = wall.Shape.Volume
        self.assertGreater(initial_wall_volume, 0, "Wall should have a positive volume initially.")

        # Create the original window's sketch
        window_width = 800.0
        window_height = 1200.0
        original_sketch = self._create_sketch_with_wires(
            "OriginalWinSketch", [(0, 0, window_width, window_height)]
        )

        # Orient the sketch to be vertical before creating the window
        # This mimics the behavior of the interactive Arch_Window tool.
        original_sketch.Placement.Rotation = FreeCAD.Rotation(FreeCAD.Vector(1, 0, 0), 90)
        self.document.recompute()

        # Create the original window from the now-oriented sketch
        original_window = Arch.makeWindow(baseobj=original_sketch, name="OriginalWindow")
        original_window.Width = window_width  # Explicitly set properties
        original_window.Height = window_height
        self.document.recompute()
        self.assertFalse(
            original_window.Shape.isNull(), "Original window shape should not be null."
        )

        # Create the clone
        cloned_window = Draft.clone(original_window)
        cloned_window.Label = "ClonedWindow"

        # Position the clone inside the wall
        clone_x = 1100.0  # Center of the clone in the wall's length
        clone_y = wall_thickness / 2  # Center of the clone in the wall's thickness
        clone_z = 800.0  # Sill height
        cloned_window.Placement.Base = FreeCAD.Vector(clone_x, clone_y, clone_z)
        self.document.recompute()

        self.assertIsNotNone(
            cloned_window.CloneOf, "Cloned window should have CloneOf property set."
        )
        self.assertEqual(
            cloned_window.CloneOf, original_window, "CloneOf should point to the original window."
        )
        self.assertAlmostEqual(cloned_window.Shape.Volume, original_window.Shape.Volume, delta=1e-5)

        # Add the clone to the wall's hosts
        Arch.addComponents(cloned_window, host=wall)
        self.document.recompute()

        self.assertIn(
            wall, cloned_window.Hosts, "Wall should be in the cloned window's Hosts list."
        )

        final_wall_volume = wall.Shape.Volume
        self.assertLess(
            final_wall_volume,
            initial_wall_volume,
            "Wall volume should have decreased after hosting the cloned window.",
        )

        expected_removed_volume = window_width * window_height * wall.Width.Value
        self.assertAlmostEqual(
            initial_wall_volume - final_wall_volume,
            expected_removed_volume,
            delta=1e-5,
            msg="The volume removed from the wall by the cloned window is incorrect.",
        )

    def test_addComponents_window_to_wall(self):
        """
        Tests the Arch.addComponents function for adding a window to a wall.
        Verifies that the Hosts, InList, and OutList properties are correctly
        updated, and that the geometric opening is created.
        """
        self.printTestMessage("Testing Arch.addComponents for window-wall hosting...")

        # Create the wall and window
        wall_base = Draft.makeLine(FreeCAD.Vector(0, 0, 0), FreeCAD.Vector(3000, 0, 0))
        wall = Arch.makeWall(wall_base, width=200, height=2500, name="HostWall")
        self.document.recompute()
        initial_wall_volume = wall.Shape.Volume

        window_width = 800.0
        window_height = 1200.0
        window_sketch = self._create_sketch_with_wires(
            "WindowSketchForAdd", [(0, 0, window_width, window_height)]
        )
        window_sketch.Placement.Rotation = FreeCAD.Rotation(FreeCAD.Vector(1, 0, 0), 90)
        window_sketch.Placement.Base = FreeCAD.Vector(1100, 100, 800)  # Position it within the wall
        self.document.recompute()

        window = Arch.makeWindow(baseobj=window_sketch, name="HostedWindow")
        window.Width = window_width
        window.Height = window_height
        self.document.recompute()

        # Initial state verification
        self.assertEqual(len(window.Hosts), 0, "Window.Hosts should be empty initially.")
        self.assertNotIn(window, wall.InList, "Window should not be in wall's InList initially.")

        # Add the window to the wall
        Arch.addComponents(window, wall)
        self.document.recompute()

        # Check all relationships and the final geometry
        # Property assertions
        self.assertIn(wall, window.Hosts, "Wall should be in window.Hosts after addComponents.")
        self.assertEqual(len(window.Hosts), 1, "Window.Hosts should contain exactly one host.")

        # Dependency graph assertions
        self.assertIn(window, wall.InList, "Window should be in wall's InList after hosting.")
        self.assertIn(wall, window.OutList, "Wall should be in window's OutList after hosting.")

        # Negative assertion
        self.assertNotIn(
            window,
            wall.Subtractions,
            "Window should not be in wall's Subtractions list for this hosting mechanism.",
        )

        # Geometric assertion
        final_wall_volume = wall.Shape.Volume
        self.assertLess(
            final_wall_volume,
            initial_wall_volume,
            "Wall volume should decrease after hosting the window.",
        )

        expected_removed_volume = window_width * window_height * wall.Width.Value
        self.assertAlmostEqual(
            initial_wall_volume - final_wall_volume,
            expected_removed_volume,
            delta=1e-5,
            msg="The volume removed from the wall is incorrect.",
        )

    def _create_sketch_with_wires(
        self, name: str, wire_definitions: list[tuple[float, float, float, float]]
    ) -> "FreeCAD.DocumentObject":
        """
        Helper to create a sketch with one or more specified rectangular wires.

        Each rectangle is defined in the sketch's local XY plane. The sketch
        is created in the current document (`self.doc`). After adding all
        geometry and basic coincident constraints for each rectangle, the
        document is recomputed.

        Parameters
        ----------
        name : str
            The name for the new Sketcher.SketchObject.
        wire_definitions : list[tuple[float, float, float, float]]
            A list where each element defines one rectangular wire.
            Each tuple within the list should be in the format `(x, y, w, h)`:
            - `x` (float): The x-coordinate of the bottom-left corner of the rectangle
                           in the sketch's local coordinate system.
            - `y` (float): The y-coordinate of the bottom-left corner of the rectangle
                           in the sketch's local coordinate system.
            - `w` (float): The width of the rectangle (along the sketch's local X-axis).
            - `h` (float): The height of the rectangle (along the sketch's local Y-axis).
            For example, `[(0, 0, 100, 200)]` creates one 100x200 rectangle at the
            sketch origin. `[(0,0,100,100), (10,10,80,80)]` creates two concentric
            rectangles if used as outer and inner boundaries. The order of wires
            in this list corresponds to `Wire0`, `Wire1`, etc., when used by
            Arch objects. The sketch is recomputed and returned.

        Returns
        -------
        FreeCAD.DocumentObject
            The created Sketcher object (specifically, an object of type "Sketcher::SketchObject").
        """
        sketch: FreeCAD.DocumentObject = self.document.addObject("Sketcher::SketchObject", name)

        for i, (x, y, w, h) in enumerate(wire_definitions):
            sketch.addGeometry(
                Part.LineSegment(FreeCAD.Vector(x, y, 0), FreeCAD.Vector(x + w, y, 0))
            )
            sketch.addGeometry(
                Part.LineSegment(FreeCAD.Vector(x + w, y, 0), FreeCAD.Vector(x + w, y + h, 0))
            )
            sketch.addGeometry(
                Part.LineSegment(FreeCAD.Vector(x + w, y + h, 0), FreeCAD.Vector(x, y + h, 0))
            )
            sketch.addGeometry(
                Part.LineSegment(FreeCAD.Vector(x, y + h, 0), FreeCAD.Vector(x, y, 0))
            )
            base_idx = i * 4
            sketch.addConstraint(Sketcher.Constraint("Coincident", base_idx, 2, base_idx + 1, 1))
            sketch.addConstraint(
                Sketcher.Constraint("Coincident", base_idx + 1, 2, base_idx + 2, 1)
            )
            sketch.addConstraint(
                Sketcher.Constraint("Coincident", base_idx + 2, 2, base_idx + 3, 1)
            )
            sketch.addConstraint(Sketcher.Constraint("Coincident", base_idx + 3, 2, base_idx, 1))

        self.document.recompute()

        return sketch

    def _create_sketch_with_named_constraints(
        self, name: str, initial_width: float, initial_height: float
    ) -> "FreeCAD.DocumentObject":
        """
        Helper to create a rectangular sketch with "Width" and "Height" named constraints.

        The sketch is created in the current document (`self.doc`) on the
        default XY plane. It consists of a single rectangle defined by four
        line segments, with its bottom-left corner at (0,0,0) in the
        sketch's local coordinates.

        Parameters
        ----------
        name : str
            The name for the new Sketcher.SketchObject.
        initial_width : float
            The initial value for the "Width" constraint (along sketch's X-axis).
        initial_height : float
            The initial value for the "Height" constraint (along sketch's Y-axis).

        Returns
        -------
        FreeCAD.DocumentObject
            The created Sketcher object (specifically, an object of type "Sketcher::SketchObject").
        """
        sketch: FreeCAD.DocumentObject = self.document.addObject("Sketcher::SketchObject", name)

        sketch.addGeometry(
            Part.LineSegment(FreeCAD.Vector(0, 0, 0), FreeCAD.Vector(initial_width, 0, 0)), False
        )
        sketch.addGeometry(
            Part.LineSegment(
                FreeCAD.Vector(initial_width, 0, 0),
                FreeCAD.Vector(initial_width, initial_height, 0),
            ),
            False,
        )
        sketch.addGeometry(
            Part.LineSegment(
                FreeCAD.Vector(initial_width, initial_height, 0),
                FreeCAD.Vector(0, initial_height, 0),
            ),
            False,
        )
        sketch.addGeometry(
            Part.LineSegment(FreeCAD.Vector(0, initial_height, 0), FreeCAD.Vector(0, 0, 0)), False
        )
        sketch.addConstraint(Sketcher.Constraint("Coincident", 0, 2, 1, 1))
        sketch.addConstraint(Sketcher.Constraint("Coincident", 1, 2, 2, 1))
        sketch.addConstraint(Sketcher.Constraint("Coincident", 2, 2, 3, 1))
        sketch.addConstraint(Sketcher.Constraint("Coincident", 3, 2, 0, 1))
        sketch.addConstraint(Sketcher.Constraint("Horizontal", 0))
        sketch.addConstraint(Sketcher.Constraint("Vertical", 1))
        sketch.addConstraint(Sketcher.Constraint("Horizontal", 2))
        sketch.addConstraint(Sketcher.Constraint("Vertical", 3))

        sketch.addConstraint(Sketcher.Constraint("DistanceX", 0, 1, 0, 2, initial_width))
        sketch.renameConstraint(sketch.ConstraintCount - 1, "Width")
        sketch.addConstraint(Sketcher.Constraint("DistanceY", 1, 1, 1, 2, initial_height))
        sketch.renameConstraint(sketch.ConstraintCount - 1, "Height")

        self.document.recompute()

        return sketch



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2019 sliptonic <shopinthewoods@gmail.com>               *
# *                 2025 Samuel Abels <knipknap@gmail.com>                  *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import Path
import Path.Base.Util as PathUtil
import json
import uuid
import pathlib
from abc import ABC
from itertools import chain
from lazy_loader.lazy_loader import LazyLoader
from typing import Any, List, Optional, Tuple, Type, Union, Mapping, cast
from PySide.QtCore import QT_TRANSLATE_NOOP
from Path.Base.Generator import toolchange
from ...docobject import DetachedDocumentObject
from ...assets.asset import Asset
from ...camassets import cam_assets
from ...shape import ToolBitShape, ToolBitShapeCustom, ToolBitShapeIcon
from ..util import to_json, format_value
from ..migration import ParameterAccessor, migrate_parameters


ToolBitView = LazyLoader("Path.Tool.toolbit.ui.view", globals(), "Path.Tool.toolbit.ui.view")


class ToolBitRecomputeObserver:
    """Document observer that triggers queued visual updates after recompute completes."""

    def __init__(self, toolbit_proxy):
        self.toolbit_proxy = toolbit_proxy

    def slotRecomputedDocument(self, doc):
        """Called when document recompute is finished."""
        # Check if the toolbit object is still valid
        try:
            obj_doc = self.toolbit_proxy.obj.Document
        except ReferenceError:
            # Object has been deleted or does not exist, nothing to do
            return

        # Only process updates for the correct document
        if doc != obj_doc:
            return

        # Process any queued visual updates
        if self.toolbit_proxy and hasattr(self.toolbit_proxy, "_process_queued_visual_update"):
            Path.Log.debug("Document recompute finished, processing queued visual update")
            self.toolbit_proxy._process_queued_visual_update()


PropertyGroupShape = "Shape"

if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


class ToolBit(Asset, ABC):
    asset_type: str = "toolbit"
    SHAPE_CLASS: Type[ToolBitShape]  # Abstract class attribute

    def __init__(self, tool_bit_shape: ToolBitShape, id: Optional[str] = None):
        Path.Log.track("ToolBit __init__ called")
        self.id = id if id is not None else str(uuid.uuid4())
        self.obj = DetachedDocumentObject()
        self.obj.Proxy = self
        self._tool_bit_shape: ToolBitShape = tool_bit_shape
        self._in_update = False

        self._create_base_properties()
        self.obj.ToolBitID = self.get_id()
        self.obj.ShapeID = tool_bit_shape.get_id()
        self.obj.ShapeType = tool_bit_shape.name
        self.obj.Label = tool_bit_shape.label or f"New {tool_bit_shape.name}"

        # Initialize properties
        self._update_tool_properties()

    def __eq__(self, other):
        """Compare ToolBit objects based on their unique ID."""
        if not isinstance(other, ToolBit):
            return False
        return self.id == other.id

    @staticmethod
    def _find_subclass_for_shape(shape: ToolBitShape) -> Type["ToolBit"]:
        """
        Finds the appropriate ToolBit subclass for a given ToolBitShape instance.
        """
        for subclass in ToolBit.__subclasses__():
            if isinstance(shape, subclass.SHAPE_CLASS):
                return subclass
        raise ValueError(f"No ToolBit subclass found for shape {type(shape).__name__}")

    @classmethod
    def from_dict(cls, attrs: Mapping, shallow: bool = False) -> "ToolBit":
        """
        Creates and populates a ToolBit instance from a dictionary.
        """
        # Find the shape ID.
        shape_id = pathlib.Path(
            str(attrs.get("shape", ""))
        ).stem  # backward compatibility. used to be a filename
        if not shape_id:
            raise ValueError("ToolBit dictionary is missing 'shape' key")

        # Try to find the shape type. Default to Unknown if necessary.
        if "shape" in attrs and "shape-type" not in attrs:
            attrs["shape-type"] = attrs["shape"]
        shape_type = attrs.get("shape-type")
        shape_class = ToolBitShape.get_shape_class_from_id(shape_id, shape_type)
        if not shape_class:
            Path.Log.debug(
                f"Failed to find usable shape for ID '{shape_id}'"
                f" (shape type {shape_type}). Falling back to 'Unknown'"
            )
            shape_class = ToolBitShapeCustom

        # Create a ToolBitShape instance.
        if not shallow:  # Shallow means: skip loading of child assets
            shape_asset_uri = ToolBitShape.resolve_name(shape_id)
            try:
                tool_bit_shape = cast(ToolBitShape, cam_assets.get(shape_asset_uri))
            except FileNotFoundError:
                Path.Log.debug(f"ToolBit.from_dict: Shape asset {shape_asset_uri} not found.")
                # Rely on the fallback below
            else:
                return cls.from_shape(tool_bit_shape, attrs, id=attrs.get("id"))

        # Ending up here means we either could not load the shape asset,
        # or we are in shallow mode and do not want to load it.
        # Create a shape instance from scratch as a "placeholder".
        params = attrs.get("parameter", {})
        tool_bit_shape = shape_class(shape_id, **params)
        Path.Log.debug(
            f"ToolBit.from_dict: created shape instance {tool_bit_shape.name}"
            f" from {shape_id}. Uri: {tool_bit_shape.get_uri()}"
        )

        # Now that we have a shape, create the toolbit instance.
        return cls.from_shape(tool_bit_shape, attrs, id=attrs.get("id"))

    @classmethod
    def from_shape(
        cls,
        tool_bit_shape: ToolBitShape,
        attrs: Mapping,
        id: Optional[str] = None,
    ) -> "ToolBit":
        selected_toolbit_subclass = cls._find_subclass_for_shape(tool_bit_shape)
        toolbit = selected_toolbit_subclass(tool_bit_shape, id=id)
        toolbit.label = attrs.get("name") or tool_bit_shape.label

        # Get params and attributes.
        params = attrs.get("parameter", {})
        attr = attrs.get("attribute", {})

        # Filter parameters if method exists
        if (
            hasattr(tool_bit_shape.__class__, "filter_parameters")
            and callable(getattr(tool_bit_shape.__class__, "filter_parameters"))
            and isinstance(params, dict)
        ):
            params = tool_bit_shape.__class__.filter_parameters(params)

        # Update parameters.
        for param_name, param_value in params.items():
            tool_bit_shape.set_parameter(param_name, param_value)
            if hasattr(toolbit.obj, param_name):
                PathUtil.setProperty(toolbit.obj, param_name, param_value)

        # Update attributes; the separation between parameters and attributes
        # is currently not well defined, so for now we add them to the
        # ToolBitShape and the DocumentObject.
        # Discussion: https://github.com/FreeCAD/FreeCAD/issues/21722
        for attr_name, attr_value in attr.items():
            tool_bit_shape.set_parameter(attr_name, attr_value)
            if hasattr(toolbit.obj, attr_name):
                PathUtil.setProperty(toolbit.obj, attr_name, attr_value)
            else:
                Path.Log.debug(
                    f"ToolBit {id} Attribute '{attr_name}' not found on"
                    f" {selected_toolbit_subclass.__name__} ({tool_bit_shape})"
                    f" '{toolbit.obj.Label}'. Skipping."
                )

        toolbit._update_tool_properties()
        return toolbit

    @classmethod
    def from_shape_id(cls, shape_id: str, label: Optional[str] = None) -> "ToolBit":
        """
        Creates and populates a ToolBit instance from a shape ID.
        """
        attrs = {"shape": shape_id, "name": label}
        return cls.from_dict(attrs)

    @classmethod
    def from_file(cls, path: Union[str, pathlib.Path]) -> "ToolBit":
        """
        Creates and populates a ToolBit instance from a .fctb file.
        """
        path = pathlib.Path(path)
        with path.open("r") as fp:
            attrs_map = json.load(fp)
        return cls.from_dict(attrs_map)

    @property
    def label(self) -> str:
        return self.obj.Label

    @label.setter
    def label(self, label: str):
        self.obj.Label = label

    def get_shape_name(self) -> str:
        """Returns the shape name of the tool bit."""
        return self._tool_bit_shape.name

    def set_shape_name(self, name: str):
        """Sets the shape name of the tool bit."""
        self._tool_bit_shape.name = name

    @property
    def summary(self) -> str:
        """
        To be overridden by subclasses to provide a better summary
        including parameter values. Used as "subtitle" for the tool
        in the UI.

        Example: "3.2 mm endmill, 4-flute, 8 mm cutting edge"
        """
        return self.get_shape_name()

    def _create_base_properties(self):
        # Create the properties in the Base group.
        if not hasattr(self.obj, "ShapeID"):
            self.obj.addProperty(
                "App::PropertyString",
                "ShapeID",
                "Base",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "The unique ID of the tool shape (.fcstd)",
                ),
            )
        if not hasattr(self.obj, "ShapeType"):
            self.obj.addProperty(
                "App::PropertyEnumeration",
                "ShapeType",
                "Base",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "The tool shape type",
                ),
            )
            names = [c.name for c in ToolBitShape.__subclasses__()]
            self.obj.ShapeType = names
            self.obj.ShapeType = ToolBitShapeCustom.name
        if not hasattr(self.obj, "BitBody"):
            self.obj.addProperty(
                "App::PropertyLink",
                "BitBody",
                "Base",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "The parametrized body representing the tool bit",
                ),
            )
        if not hasattr(self.obj, "ToolBitID"):
            self.obj.addProperty(
                "App::PropertyString",
                "ToolBitID",
                "Base",
                QT_TRANSLATE_NOOP("App::Property", "The unique ID of the toolbit"),
            )

        # 0 = read/write, 1 = read only, 2 = hide
        self.obj.setEditorMode("ShapeID", 1)
        self.obj.setEditorMode("ShapeType", 1)
        self.obj.setEditorMode("ToolBitID", 1)
        self.obj.setEditorMode("BitBody", 2)
        self.obj.setEditorMode("Shape", 2)

        # Create the ToolBit properties that are shared by all tool bits
        if not hasattr(self.obj, "SpindleDirection"):
            self.obj.addProperty(
                "App::PropertyEnumeration",
                "SpindleDirection",
                "Attributes",
                QT_TRANSLATE_NOOP("App::Property", "Direction of spindle rotation"),
            )
            self.obj.SpindleDirection = ["Forward", "Reverse", "None"]
            self.obj.SpindleDirection = "Forward"  # Default value
        if not hasattr(self.obj, "Material"):
            self.obj.addProperty(
                "App::PropertyEnumeration",
                "Material",
                "Attributes",
                QT_TRANSLATE_NOOP("App::Property", "Tool material"),
            )
            self.obj.Material = ["HSS", "Carbide"]
            self.obj.Material = "HSS"  # Default value

    def get_id(self) -> str:
        """Returns the unique ID of the tool bit."""
        return self.id

    def set_id(self, id: str = None):
        self.id = id if id is not None else str(uuid.uuid4())

    def _promote_toolbit(self):
        """
        Updates the toolbit properties for backward compatibility.
        Ensure obj.ShapeID and obj.ToolBitID are set, handling legacy cases.
        Also promotes embedded toolbits to correct shape type if needed.
        """
        Path.Log.track(f"Promoting tool bit {self.obj.Label}")

        # Ensure ShapeID is set (handling legacy BitShape/ShapeName)
        name = None
        if hasattr(self.obj, "ShapeID") and self.obj.ShapeID:
            name = self.obj.ShapeID
        elif hasattr(self.obj, "ShapeFile") and self.obj.ShapeFile:
            name = pathlib.Path(self.obj.ShapeFile).stem
        elif hasattr(self.obj, "BitShape") and self.obj.BitShape:
            name = pathlib.Path(self.obj.BitShape).stem
        elif hasattr(self.obj, "ShapeName") and self.obj.ShapeName:
            name = pathlib.Path(self.obj.ShapeName).stem
        if name is None:
            raise ValueError("ToolBit is missing a shape ID")

        uri = ToolBitShape.resolve_name(name)
        if uri is None:
            raise ValueError(f"Failed to identify ID of ToolBit from '{name}'")
        self.obj.ShapeID = uri.asset_id

        # Ensure ShapeType is set
        thetype = None
        if hasattr(self.obj, "ShapeType") and self.obj.ShapeType:
            thetype = self.obj.ShapeType
        elif hasattr(self.obj, "ShapeFile") and self.obj.ShapeFile:
            thetype = pathlib.Path(self.obj.ShapeFile).stem
        elif hasattr(self.obj, "BitShape") and self.obj.BitShape:
            thetype = pathlib.Path(self.obj.BitShape).stem
        elif hasattr(self.obj, "ShapeName") and self.obj.ShapeName:
            thetype = pathlib.Path(self.obj.ShapeName).stem
        if thetype is None:
            raise ValueError("ToolBit is missing a shape type")

        shape_class = ToolBitShape.get_subclass_by_name(thetype)
        if shape_class is None:
            raise ValueError(f"Failed to identify shape of ToolBit from '{thetype}'")
        self.obj.ShapeType = shape_class.name

        # Promote embedded toolbits to correct shape type if still Custom
        if self.obj.ShapeType == "Custom":
            shape_id = getattr(self.obj, "ShapeID", None)
            if shape_id:
                shape_class = ToolBitShape.get_subclass_by_name(shape_id)
                if shape_class and shape_class.name != "Custom":
                    self.obj.ShapeType = shape_class.name
                    self._tool_bit_shape = shape_class(shape_id)
                    Path.Log.info(
                        f"Promoted embedded toolbit '{self.obj.Label}' to shape '{shape_class.name}' via ShapeID"
                    )
        # Ensure ToolBitID is set
        if hasattr(self.obj, "File"):
            self.id = pathlib.Path(self.obj.File).stem
        self.obj.ToolBitID = self.id
        Path.Log.debug(f"Set ToolBitID to {self.obj.ToolBitID}")

        # Update SpindleDirection:
        # Old tools may still have "CCW", "CW", "Off", "None".
        # New tools use "None", "Forward", "Reverse".
        normalized_direction = old_direction = self.obj.SpindleDirection

        if isinstance(old_direction, str):
            lower_direction = old_direction.lower()
            if lower_direction in ("none", "off"):
                normalized_direction = "None"
            elif lower_direction in ("cw", "forward"):
                normalized_direction = "Forward"
            elif lower_direction in ("ccw", "reverse"):
                normalized_direction = "Reverse"

        self.obj.SpindleDirection = ["Forward", "Reverse", "None"]
        self.obj.SpindleDirection = normalized_direction
        if old_direction != normalized_direction:
            Path.Log.info(
                f"Promoted tool bit {self.obj.Label}: SpindleDirection from {old_direction} to {self.obj.SpindleDirection}"
            )

        # Drop legacy properties.
        legacy = "ShapeFile", "File", "BitShape", "ShapeName"
        for name in legacy:
            if hasattr(self.obj, name):
                value = getattr(self.obj, name)
                self.obj.removeProperty(name)
                Path.Log.debug(f"Removed obsolete property '{name}' ('{value}').")

        # Get the schema properties from the current shape
        shape_cls = ToolBitShape.get_subclass_by_name(self.obj.ShapeType)
        if not shape_cls:
            raise ValueError(f"Failed to find shape class named '{self.obj.ShapeType}'")
        shape_schema_props = shape_cls.schema().keys()

        # Move properties that are part of the shape schema to the "Shape" group
        for prop_name in self.obj.PropertiesList:
            if (
                self.obj.getGroupOfProperty(prop_name) == PropertyGroupShape
                or prop_name not in shape_schema_props
            ):
                continue
            try:
                Path.Log.debug(f"Moving property '{prop_name}' to group '{PropertyGroupShape}'")

                # Get property details before removing
                prop_type = self.obj.getTypeIdOfProperty(prop_name)
                prop_doc = self.obj.getDocumentationOfProperty(prop_name)
                prop_value = self.obj.getPropertyByName(prop_name)

                # Remove the property
                self.obj.removeProperty(prop_name)

                # Add the property back to the Shape group
                self.obj.addProperty(prop_type, prop_name, PropertyGroupShape, prop_doc)
                self._in_update = True  # Prevent onChanged from running
                PathUtil.setProperty(self.obj, prop_name, prop_value)
                Path.Log.info(f"Moved property '{prop_name}' to group '{PropertyGroupShape}'")
            except Exception as e:
                Path.Log.error(
                    f"Failed to move property '{prop_name}' to group '{PropertyGroupShape}': {e}"
                )
                raise
            finally:
                self._in_update = False

    def onDocumentRestored(self, obj):
        Path.Log.track(obj.Label)

        # Assign self.obj to the restored object
        self.obj = obj
        self.obj.Proxy = self
        if not hasattr(self, "id"):
            self.id = str(uuid.uuid4())
            Path.Log.debug(
                f"Assigned new id {self.id} for ToolBit {obj.Label} during document restore"
            )

        # Our constructor previously created the base properties in the
        # DetachedDocumentObject, which was now replaced.
        # So here we need to ensure to set them up in the new (real) DocumentObject
        # as well.
        self._create_base_properties()
        self._promote_toolbit()

        # Get the shape instance based on the ShapeType. We try two approaches
        # to find the shape and shape class:
        #   1. If the asset with the given type exists, use that.
        #   2. Otherwise create a new empty instance
        shape_uri = ToolBitShape.resolve_name(self.obj.ShapeType)
        try:
            # Best case: we directly find the shape file in our assets.
            self._tool_bit_shape = cast(ToolBitShape, cam_assets.get(shape_uri))
        except FileNotFoundError:
            # Otherwise, try to at least identify the type of the shape.
            shape_class = ToolBitShape.get_subclass_by_name(shape_uri.asset_id)
            if not shape_class:
                raise ValueError(
                    "Failed to identify class of ToolBitShape from name "
                    f"'{self.obj.ShapeType}' (asset id {shape_uri.asset_id})"
                )
            self._tool_bit_shape = shape_class(shape_uri.asset_id)

        # If BitBody exists and is in a different document after document restore,
        # it means a shallow copy occurred. We need to re-initialize the visual
        # representation and properties to ensure a deep copy of the BitBody
        # and its properties.
        # Only re-initialize properties from shape if not restoring from file
        if self.obj.BitBody and self.obj.BitBody.Document != self.obj.Document:
            Path.Log.debug(
                f"onDocumentRestored: Re-initializing BitBody for {self.obj.Label} after copy"
            )
            self._update_visual_representation()

        # Ensure the correct ViewProvider is attached during document restore,
        # because some legacy fcstd files may still have references to old view
        # providers.
        if hasattr(self.obj, "ViewObject") and self.obj.ViewObject:
            if hasattr(self.obj.ViewObject, "Proxy") and not isinstance(
                self.obj.ViewObject.Proxy, ToolBitView.ViewProvider
            ):
                Path.Log.debug(f"onDocumentRestored: Attaching ViewProvider for {self.obj.Label}")
                ToolBitView.ViewProvider(self.obj.ViewObject, "ToolBit")

        # Migrate legacy parameters using unified accessor
        migrate_parameters(ParameterAccessor(obj))

        # Filter parameters if method exists (removes FlatRadius from obj)
        filter_func = getattr(self._tool_bit_shape.__class__, "filter_parameters", None)
        if callable(filter_func):
            # Only filter if FlatRadius is present
            if "FlatRadius" in self.obj.PropertiesList:
                try:
                    self.obj.removeProperty("FlatRadius")
                    Path.Log.info(f"Filtered out FlatRadius for {self.obj.Label}")
                except Exception as e:
                    Path.Log.error(f"Failed to remove FlatRadius for {self.obj.Label}: {e}")

        # Copy properties from the restored object to the ToolBitShape.
        for name, item in self._tool_bit_shape.schema().items():
            if name in self.obj.PropertiesList:
                value = self.obj.getPropertyByName(name)
                self._tool_bit_shape.set_parameter(name, value)

        # Ensure property state is correct after restore.
        self._update_tool_properties()

    def attach_to_doc(
        self, doc: FreeCAD.Document, label: Optional[str] = None
    ) -> FreeCAD.DocumentObject:
        """
        Creates a new FreeCAD DocumentObject in the given document and attaches
        this ToolBit instance to it.
        """
        label = label or self.label or self._tool_bit_shape.label
        tool_doc_obj = doc.addObject("Part::FeaturePython", label)
        self.attach_to_obj(tool_doc_obj, label=label)
        return tool_doc_obj

    def attach_to_obj(self, tool_doc_obj: FreeCAD.DocumentObject, label: Optional[str] = None):
        """
        Attaches the ToolBit instance to an existing FreeCAD DocumentObject.

        Transfers properties from the internal DetachedDocumentObject to the
        tool_doc_obj and updates the visual representation.
        """
        if not isinstance(self.obj, DetachedDocumentObject):
            Path.Log.warning(
                f"ToolBit {self.obj.Label} is already attached to a "
                "DocumentObject. Skipping attach_to_obj."
            )
            return

        Path.Log.track(f"Attaching ToolBit to {tool_doc_obj.Label}")

        temp_obj = self.obj
        self.obj = tool_doc_obj
        self.obj.Proxy = self
        if FreeCAD.GuiUp:
            ToolBitView.ViewProvider(self.obj.ViewObject, "ToolBit")

        self._create_base_properties()

        # Transfer property values from the detached object to the real object
        self._suppress_visual_update = True
        temp_obj.copy_to(self.obj)

        # Ensure label is set
        self.obj.Label = label or self.label or self._tool_bit_shape.label

        # Update the visual representation now that it's attached
        self._update_tool_properties()
        self._suppress_visual_update = False
        self._update_visual_representation()

    def onChanged(self, obj, prop):
        Path.Log.track(obj.Label, prop)
        # Avoid acting during document restore or internal updates
        if "Restore" in obj.State:
            return

        if getattr(self, "_suppress_visual_update", False):
            return

        if hasattr(self, "_in_update") and self._in_update:
            Path.Log.debug(f"Skipping onChanged for {obj.Label} due to active update.")
            return

        # We only care about updates that affect the Shape
        if obj.getGroupOfProperty(prop) != PropertyGroupShape:
            return

        self._in_update = True
        try:
            new_value = obj.getPropertyByName(prop)
            Path.Log.debug(
                f"Shape parameter '{prop}' changed to {new_value}. "
                f"Queuing visual representation update."
            )
            self._tool_bit_shape.set_parameter(prop, new_value)
            self._queue_visual_update()
        finally:
            self._in_update = False

    def onDelete(self, obj, arg2=None):
        Path.Log.track(obj.Label)
        # Clean up any pending observer
        if hasattr(self, "_recompute_observer"):
            FreeCAD.removeDocumentObserver(self._recompute_observer)
            del self._recompute_observer
        self._removeBitBody()
        obj.Document.removeObject(obj.Name)

    def _removeBitBody(self):
        if self.obj.BitBody:
            self.obj.BitBody.removeObjectsFromDocument()
            self.obj.Document.removeObject(self.obj.BitBody.Name)
            self.obj.BitBody = None

    def _setupProperty(self, prop, orig):
        # extract property parameters and values so it can be copied
        val = orig.getPropertyByName(prop)
        typ = orig.getTypeIdOfProperty(prop)
        grp = orig.getGroupOfProperty(prop)
        dsc = orig.getDocumentationOfProperty(prop)

        self.obj.addProperty(typ, prop, grp, dsc)
        if "App::PropertyEnumeration" == typ:
            setattr(self.obj, prop, orig.getEnumerationsOfProperty(prop))
        self.obj.setEditorMode(prop, 1)
        PathUtil.setProperty(self.obj, prop, val)

    def _get_props(self, group: Optional[Union[str, Tuple[str, ...]]] = None) -> List[str]:
        """
        Returns a list of property names from the given group(s) for the object.
        Returns all groups if the group argument is None.
        """
        props_in_group = []
        # Use PropertiesList to get all property names
        for prop in self.obj.PropertiesList:
            prop_group = self.obj.getGroupOfProperty(prop)
            if group is None:
                props_in_group.append(prop)
            elif isinstance(group, str) and prop_group == group:
                props_in_group.append(prop)
            elif isinstance(group, tuple) and prop_group in group:
                props_in_group.append(prop)
        return props_in_group

    def get_property(self, name: str):
        return self.obj.getPropertyByName(name)

    def get_property_str(
        self, name: str, default: str | None = None, precision: int | None = None
    ) -> str | None:
        value = self.get_property(name)
        return format_value(value, precision=precision) if value else default

    def set_property(self, name: str, value: Any):
        return self.obj.setPropertyByName(name, value)

    def get_property_label_from_name(self, name: str):
        return self.obj.getPropertyByName

    def get_icon(self) -> Optional[ToolBitShapeIcon]:
        """
        Retrieves the thumbnail data for the tool bit shape, as
        taken from the explicit SVG or PNG, if the shape has one.
        """
        if self._tool_bit_shape:
            return self._tool_bit_shape.get_icon()
        return None

    def get_thumbnail(self) -> Optional[bytes]:
        """
        Retrieves the thumbnail data for the tool bit shape in PNG format,
        as embedded in the shape file.
        Fallback to the icon from get_icon() (converted to PNG)
        """
        if not self._tool_bit_shape:
            return None
        png_data = self._tool_bit_shape.get_thumbnail()
        if png_data:
            return png_data
        icon = self.get_icon()
        if icon:
            return icon.get_png()
        return None

    def _remove_properties(self, group, prop_names):
        for name in prop_names:
            if hasattr(self.obj, name):
                if self.obj.getGroupOfProperty(name) == group:
                    try:
                        self.obj.removeProperty(name)
                        Path.Log.debug(f"Removed property: {group}.{name}")
                    except Exception as e:
                        Path.Log.error(f"Failed removing property '{group}.{name}': {e}")
            else:
                Path.Log.warning(f"'{group}.{name}' failed to remove property, not found")

    def _update_tool_properties(self):
        """
        Initializes or updates the tool bit's properties based on the current
        _tool_bit_shape. Adds/updates shape parameters, removes obsolete shape
        parameters, and updates the edit state of them.
        Does not handle updating the visual representation.
        """
        Path.Log.track(self.obj.Label)

        # 1. Add/Update properties for the new shape
        for name, item in self._tool_bit_shape.schema().items():
            docstring = item[0]
            prop_type = item[1]

            if not prop_type:
                Path.Log.error(
                    f"No property type for parameter '{name}' in shape "
                    f"'{self._tool_bit_shape.name}'. Skipping."
                )
                continue

            # Add new property
            if not hasattr(self.obj, name):
                self.obj.addProperty(prop_type, name, "Shape", docstring)
                Path.Log.debug(f"Added new shape property: {name}")

            # Ensure editor mode is correct
            self.obj.setEditorMode(name, 0)

            try:
                value = self._tool_bit_shape.get_parameter(name)
            except KeyError:
                continue  # Retain existing property value.

            # Conditional to avoid unnecessary migration warning when called
            # from onDocumentRestored.
            if value is not None and getattr(self.obj, name) != value:
                PathUtil.setProperty(self.obj, name, value)

        # 2. Add additional properties that are part of the shape,
        # but not part of the schema.
        schema_prop_names = set(self._tool_bit_shape.schema().keys())
        for name, value in self._tool_bit_shape.get_parameters().items():
            if name in schema_prop_names:
                continue
            prop_type = self._tool_bit_shape.get_parameter_type(name)
            docstring = QT_TRANSLATE_NOOP("App::Property", f"Custom property from shape: {name}")

            # Skip existing properties if they have a different type
            if hasattr(self.obj, name) and self.obj.getTypeIdOfProperty(name) != prop_type:
                Path.Log.debug(
                    f"Skipping existing property '{name}' due to type mismatch."
                    f" has type {self.obj.getTypeIdOfProperty(name)}, expected {prop_type}"
                )
                continue

            # Add the property if it does not exist
            if not hasattr(self.obj, name):
                self.obj.addProperty(prop_type, name, PropertyGroupShape, docstring)
                Path.Log.debug(f"Added custom shape property: {name} ({prop_type})")

            # Set the property value
            if value is not None and getattr(self.obj, name) != value:
                PathUtil.setProperty(self.obj, name, value)
            self.obj.setEditorMode(name, 0)

        # 3. Ensure SpindleDirection property exists and is set
        # Maybe this could be done with a global schema or added to each
        # shape schema?
        if not hasattr(self.obj, "SpindleDirection"):
            self.obj.addProperty(
                "App::PropertyEnumeration",
                "SpindleDirection",
                "Attributes",
                QT_TRANSLATE_NOOP("App::Property", "Direction of spindle rotation"),
            )
            self.obj.SpindleDirection = ["Forward", "Reverse", "None"]
            self.obj.SpindleDirection = "Forward"  # Default value

        spindle_value = self._tool_bit_shape.get_parameters().get("SpindleDirection")
        if (
            spindle_value in ("Forward", "Reverse", "None")
            and self.obj.SpindleDirection != spindle_value
        ):
            # self.obj.SpindleDirection = spindle_value
            PathUtil.setProperty(self.obj, "SpindleDirection", spindle_value)

        # 4. Ensure Material property exists and is set
        if not hasattr(self.obj, "Material"):
            self.obj.addProperty(
                "App::PropertyEnumeration",
                "Material",
                "Attributes",
                QT_TRANSLATE_NOOP("App::Property", "Tool material"),
            )
            self.obj.Material = ["HSS", "Carbide"]
            self.obj.Material = "HSS"  # Default value

        material_value = self._tool_bit_shape.get_parameters().get("Material")
        if material_value in ("HSS", "Carbide") and self.obj.Material != material_value:
            PathUtil.setProperty(self.obj, "Material", material_value)

    def _queue_visual_update(self):
        """Queue a visual update to be processed after document recompute is complete."""
        if not hasattr(self, "_visual_update_queued"):
            self._visual_update_queued = False

        if not self._visual_update_queued:
            self._visual_update_queued = True
            Path.Log.debug(f"Queuing visual update for {self.obj.Label}")

            # Set up a document observer to process the update after recompute
            self._setup_recompute_observer()

    def _setup_recompute_observer(self):
        """Set up a document observer to process queued visual updates after recompute."""
        if not hasattr(self, "_recompute_observer"):
            Path.Log.debug(f"Setting up recompute observer for {self.obj.Label}")
            self._recompute_observer = ToolBitRecomputeObserver(self)
            FreeCAD.addDocumentObserver(self._recompute_observer)

    def _process_queued_visual_update(self):
        """Process the queued visual update."""
        if hasattr(self, "_visual_update_queued") and self._visual_update_queued:
            self._visual_update_queued = False
            Path.Log.debug(f"Processing queued visual update for {self.obj.Label}")
            self._update_visual_representation()

            # Clean up the observer
            if hasattr(self, "_recompute_observer"):
                FreeCAD.removeDocumentObserver(self._recompute_observer)
                del self._recompute_observer

    def _update_visual_representation(self):
        """
        Updates the visual representation of the tool bit based on the current
        _tool_bit_shape. Creates or updates the BitBody and copies its shape
        to the main object.
        """
        if isinstance(self.obj, DetachedDocumentObject):
            return
        Path.Log.track(self.obj.Label)

        # Remove existing BitBody if it exists
        self._removeBitBody()

        try:
            # Use the shape's make_body method to create the visual representation
            body = self._tool_bit_shape.make_body(self.obj.Document)

            if not body:
                Path.Log.error(
                    f"Failed to create visual representation for shape "
                    f"'{self._tool_bit_shape.name}'"
                )
                return

            # Assign the created object to BitBody and copy its shape
            self.obj.BitBody = body
            self.obj.Shape = self.obj.BitBody.Shape  # Copy the evaluated Solid shape

            # Hide the visual representation and remove from tree
            if hasattr(self.obj.BitBody, "ViewObject") and self.obj.BitBody.ViewObject:
                self.obj.BitBody.ViewObject.Visibility = False
                self.obj.BitBody.ViewObject.ShowInTree = False

        except Exception as e:
            Path.Log.error(
                f"Failed to create visual representation using make_body for shape"
                f" '{self._tool_bit_shape.name}': {e}"
            )
            raise

        # clear the touched state since visual updates shouldn't require recompute
        self.obj.purgeTouched()

    def to_dict(self):
        """
        Returns a dictionary representation of the tool bit.

        Returns:
            A dictionary with tool bit properties, JSON-serializable.
        """
        Path.Log.track(self.obj.Label)
        attrs = {}
        attrs["version"] = 2
        attrs["id"] = self.id
        attrs["name"] = self.obj.Label
        attrs["shape"] = self._tool_bit_shape.get_id() + ".fcstd"
        attrs["shape-type"] = self._tool_bit_shape.name
        attrs["parameter"] = {}
        attrs["attribute"] = {}

        # Store all shape parameter names and attribute names
        param_names = self._tool_bit_shape.get_parameters()
        attr_props = self._get_props("Attributes")
        property_names = list(chain(param_names, attr_props))
        for name in property_names:
            value = getattr(self.obj, name, None)
            if value is None or isinstance(value, FreeCAD.DocumentObject):
                Path.Log.warning(
                    f"Excluding property '{name}' from serialization "
                    f"(type {type(value).__name__ if value is not None else 'None'}, value {value})"
                )
            try:
                serialized_value = to_json(value)
                attrs["parameter"][name] = serialized_value
            except (TypeError, ValueError) as e:
                Path.Log.warning(
                    f"Excluding property '{name}' from serialization "
                    f"(type {type(value).__name__}, value {value}): {e}"
                )

        Path.Log.debug(f"to_dict output for {self.obj.Label}: {attrs}")
        return attrs

    def __getstate__(self):
        """
        Prepare the ToolBit for pickling by excluding non-picklable attributes.

        Returns:
            A dictionary with picklable and JSON-serializable state.
        """
        Path.Log.track("ToolBit.__getstate__")
        state = {
            "id": getattr(self, "id", str(uuid.uuid4())),  # Fallback to new UUID
            "_in_update": getattr(self, "_in_update", False),  # Fallback to False
            "_obj_data": self.to_dict(),
        }

        if not getattr(self, "_tool_bit_shape", None):
            return state

        # Store minimal shape data to reconstruct _tool_bit_shape
        state["_shape_data"] = {
            "id": self._tool_bit_shape.get_id(),
            "name": self._tool_bit_shape.name,
            "parameters": {
                name: to_json(getattr(self.obj, name, None))
                for name in self._tool_bit_shape.get_parameters()
                if not isinstance(getattr(self.obj, name, None), FreeCAD.DocumentObject)
            },
        }

        return state

    def get_spindle_direction(self) -> toolchange.SpindleDirection:
        # To be safe, never allow non-rotatable shapes (such as probes) to rotate.
        if not self.can_rotate():
            return toolchange.SpindleDirection.OFF

        # Otherwise use power from defined attribute.
        if hasattr(self.obj, "SpindleDirection") and self.obj.SpindleDirection is not None:
            if self.obj.SpindleDirection.lower() in ("cw", "forward"):
                return toolchange.SpindleDirection.CW
            else:
                return toolchange.SpindleDirection.CCW

        # Default to keeping spindle off.
        return toolchange.SpindleDirection.OFF

    def can_rotate(self) -> bool:
        """
        Whether the spindle is allowed to rotate for this kind of ToolBit.
        This mostly exists as a safe-hold for probes, which should never rotate.
        """
        return True



# ***************************************************************************
# *   Copyright (c) 2023 Uwe Stöhr <uwestoehr@lyx.org>                      *
# *   Copyright (c) 2025 Mario Passaglia <mpassaglia[at]cbc.uba.ar>         *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM constraint current density document object"
__author__ = "Uwe Stöhr, Mario Passaglia"
__url__ = "https://www.freecad.org"

## @package constraint_currentdensity
#  \ingroup FEM
#  \brief constraint current density object

from FreeCAD import Base

from . import base_fempythonobject

_PropHelper = base_fempythonobject._PropHelper


class ConstraintCurrentDensity(base_fempythonobject.BaseFemPythonObject):

    Type = "Fem::ConstraintCurrentDensity"

    def __init__(self, obj):
        super().__init__(obj)

        for prop in self._get_properties():
            prop.add_to_object(obj)

    def _get_properties(self):
        prop = []

        prop.append(
            _PropHelper(
                type="App::PropertyCurrentDensity",
                name="CurrentDensity_re_1",
                group="Current Density",
                doc="Real part of current density x-component",
                value="0 A/m^2",
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyCurrentDensity",
                name="CurrentDensity_re_2",
                group="Current Density",
                doc="Real part of current density y-component",
                value="0 A/m^2",
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyCurrentDensity",
                name="CurrentDensity_re_3",
                group="Current Density",
                doc="Real part of current density z-component",
                value="0 A/m^2",
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyCurrentDensity",
                name="CurrentDensity_im_1",
                group="Current Density",
                doc="Imaginary part of current density x-component",
                value="0 A/m^2",
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyCurrentDensity",
                name="CurrentDensity_im_2",
                group="Current Density",
                doc="Imaginary part of current density y-component",
                value="0 A/m^2",
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyCurrentDensity",
                name="CurrentDensity_im_3",
                group="Current Density",
                doc="Imaginary part of current density z-component",
                value="0 A/m^2",
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyBool",
                name="EnableCurrentDensity_1",
                group="Current Density",
                doc="Enable currenty density x component",
                value=False,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyBool",
                name="EnableCurrentDensity_2",
                group="Current Density",
                doc="Enable currenty density y component",
                value=False,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyBool",
                name="EnableCurrentDensity_3",
                group="Current Density",
                doc="Enable currenty density z component",
                value=False,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyCurrentDensity",
                name="NormalCurrentDensity_re",
                group="Current Density",
                doc="Real part of current density normal to boundary",
                value="0 A/m^2",
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyCurrentDensity",
                name="NormalCurrentDensity_im",
                group="Current Density",
                doc="Imaginary part of current density normal to boundary",
                value="0 A/m^2",
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyEnumeration",
                name="Mode",
                group="Current Density",
                doc="Set current boundary condition direction mode",
                value=["Custom", "Normal"],
            )
        )

        return prop

    def onDocumentRestored(self, obj):
        # update old project with new properties
        for prop in self._get_properties():
            try:
                obj.getPropertyByName(prop.name)
            except Base.PropertyError:
                prop.add_to_object(obj)

        # enable current density properties from old properties
        try:
            obj.EnableCurrentDensity_1 = not obj.getPropertyByName(
                "CurrentDensity_re_1_Disabled"
            ) or not obj.getPropertyByName("CurrentDensity_im_1_Disabled")

            obj.EnableCurrentDensity_2 = not obj.getPropertyByName(
                "CurrentDensity_re_2_Disabled"
            ) or not obj.getPropertyByName("CurrentDensity_im_2_Disabled")

            obj.EnableCurrentDensity_3 = not obj.getPropertyByName(
                "CurrentDensity_re_3_Disabled"
            ) or not obj.getPropertyByName("CurrentDensity_im_3_Disabled")

            # remove old properties
            obj.setPropertyStatus("CurrentDensity_re_1_Disabled", "-LockDynamic")
            obj.removeProperty("CurrentDensity_re_1_Disabled")
            obj.setPropertyStatus("CurrentDensity_re_2_Disabled", "-LockDynamic")
            obj.removeProperty("CurrentDensity_re_2_Disabled")
            obj.setPropertyStatus("CurrentDensity_re_3_Disabled", "-LockDynamic")
            obj.removeProperty("CurrentDensity_re_3_Disabled")
            obj.setPropertyStatus("CurrentDensity_im_1_Disabled", "-LockDynamic")
            obj.removeProperty("CurrentDensity_im_1_Disabled")
            obj.setPropertyStatus("CurrentDensity_im_2_Disabled", "-LockDynamic")
            obj.removeProperty("CurrentDensity_im_2_Disabled")
            obj.setPropertyStatus("CurrentDensity_im_3_Disabled", "-LockDynamic")
            obj.removeProperty("CurrentDensity_im_3_Disabled")

        except Base.PropertyError:
            pass



# ***************************************************************************
# *   Copyright (c) 2013 Juergen Riegel <FreeCAD@juergen-riegel.net>        *
# *   Copyright (c) 2016 Bernd Hahnebach <bernd@bimstatik.org>              *
# *   Copyright (c) 2024 Mario Passaglia <mpassaglia[at]cbc.uba.ar>         *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM material document object"
__author__ = "Juergen Riegel, Bernd Hahnebach, Mario Passaglia"
__url__ = "https://www.freecad.org"

## @package material_common
#  \ingroup FEM
#  \brief material common object

from FreeCAD import Base, Units
import Materials

from . import base_fempythonobject

_PropHelper = base_fempythonobject._PropHelper


class MaterialCommon(base_fempythonobject.BaseFemPythonObject):
    """
    The MaterialCommon object
    """

    Type = "Fem::MaterialCommon"

    def __init__(self, obj):
        super().__init__(obj)

        for prop in self._get_properties():
            prop.add_to_object(obj)

        obj.addExtension("App::SuppressibleExtensionPython")

    def _get_properties(self):
        prop = []

        prop.append(
            _PropHelper(
                type="App::PropertyLinkSubListGlobal",
                name="References",
                group="Material",
                doc="List of material shapes",
                value=[],
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyEnumeration",
                name="Category",
                group="Material",
                doc="Material type: fluid or solid",
                value=["Solid", "Fluid"],
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyString",
                name="UUID",
                group="Material",
                doc="Material UUID",
                hidden=True,
                value="",
            )
        )

        return prop

    def onDocumentRestored(self, obj):
        # update old project with new properties
        for prop in self._get_properties():
            try:
                obj.getPropertyByName(prop.name)
            except Base.PropertyError:
                prop.add_to_object(obj)

            if prop.name == "References":
                # change References to App::PropertyLinkSubListGlobal
                prop.handle_change_type(obj, old_type="App::PropertyLinkSubList")

        # try update UUID from Material
        if not obj.UUID:
            obj.UUID = self._get_material_uuid(obj.Material)

        if not obj.hasExtension("App::SuppressibleExtensionPython"):
            obj.addExtension("App::SuppressibleExtensionPython")

    def _get_material_uuid(self, material):
        if not material:
            return ""

        material_manager = Materials.MaterialManager()

        for a_mat in material_manager.Materials:
            unmatched_item = True
            a_mat_prop = material_manager.getMaterial(a_mat).Properties
            for it in material:
                if it in a_mat_prop:
                    # first try to compare quantities
                    try:
                        unmatched_item = Units.Quantity(material[it]) != Units.Quantity(
                            a_mat_prop[it]
                        )
                    except ValueError:
                        # if there is no quantity, compare values directly
                        unmatched_item = material[it] != a_mat_prop[it]

                if unmatched_item:
                    break

            if not unmatched_item:
                # all material items are found in a_mat
                return a_mat

        return ""

        """
        Some remarks to the category. Not finished, thus to be continued.

        Following question need to be answered:
        Why use a attribute to split the object? If a new fem object is needed,
        a new fem object should be created. A new object will have an own Type
        and will be collected for the writer by this type.

        The category should not be used in writer! This would be the border.
        If an fem object has to be distinguished in writer it should be an own
        fem object.

        The category is just some helper to make it easier for the user to
        distinguish between different material categories.
        It can have own command, own icon, own make method. In material TaskPanel
        it can be distinguished which materials will be shown.

        ATM in calculix writer the Category is used. See comments in CalculiX Solver.
        """



# SPDX-License-Identifier: LGPL-2.1-or-later

# /***************************************************************************
#  *   Copyright (c) 2023 David Friedli <david[at]friedli-be.ch>             *
#  *                                                                         *
#  *   This file is part of FreeCAD.                                         *
#  *                                                                         *
#  *   FreeCAD is free software: you can redistribute it and/or modify it    *
#  *   under the terms of the GNU Lesser General Public License as           *
#  *   published by the Free Software Foundation, either version 2.1 of the  *
#  *   License, or (at your option) any later version.                       *
#  *                                                                         *
#  *   FreeCAD is distributed in the hope that it will be useful, but        *
#  *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
#  *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
#  *   Lesser General Public License for more details.                       *
#  *                                                                         *
#  *   You should have received a copy of the GNU Lesser General Public      *
#  *   License along with FreeCAD. If not, see                               *
#  *   <https://www.gnu.org/licenses/>.                                      *
#  *                                                                         *
#  **************************************************************************/

import FreeCAD
from FreeCAD import Units, Placement
from UtilsMeasure import MeasureBasePython
from PySide.QtCore import QT_TRANSLATE_NOOP


__title__ = "Measure Center of Mass Object"
__author__ = "David Friedli"
__url__ = "http://www.freecad.org"


"""
    The Measure cpp object defines a result and a placement property. The Python measure type
    adds it's own specific properties. Once the object is recomputed the parent properties are updated
    based on the specific python properties.

    We'll need some kind of interface for the measure command which exposes "parseSelection", "isValidSelection" etc.

"""


def makeMeasureCOM(name="CenterOfMass"):
    """makeMeasureCOM(name): make a CenterofMass measurement"""
    obj = FreeCAD.ActiveDocument.addObject("Measure::MeasurePython", name)
    MeasureCOM(obj)
    return obj


class MeasureCOM(MeasureBasePython):
    "The MeasureCOM object"

    def __init__(self, obj):
        obj.Proxy = self

        obj.addProperty(
            "App::PropertyLinkSubGlobal",
            "Element",
            "",
            QT_TRANSLATE_NOOP("App::Property", "Element to measure"),
            locked=True,
        )
        obj.addProperty(
            "App::PropertyPosition",
            "Result",
            "",
            QT_TRANSLATE_NOOP("App::PropertyVector", "The result location"),
            locked=True,
        )

    @classmethod
    def isValidSelection(cls, selection):
        if not len(selection) == 1:
            return False

        element = selection[0]
        ob = element["object"]
        subName = element["subName"]

        if not ob:
            return

        sub = ob.getSubObject(subName)
        if not sub:
            return

        if not hasattr(sub, "CenterOfMass"):
            return

        return True

    @classmethod
    def isPrioritySelection(cls, selection):
        return False

    @classmethod
    def getInputProps(cls):
        return ("Element",)

    def getSubject(self, obj):
        if not obj:
            return ()

        element = obj.Element
        if not element:
            return ()

        ob = element[0]
        if not ob:
            return ()
        return (ob,)

    def parseSelection(self, obj, selection):
        item = selection[0]
        o = item["object"]
        obj.Element = (o, item["subName"])

    def getResultString(self, obj):
        values = [Units.Quantity(v, Units.Length).getUserPreferred()[0] for v in obj.Result]
        return "COM\nX: {}\nY: {}\nZ: {}".format(*values)

    def execute(self, obj):
        element = obj.Element
        if not element:
            return

        ob = element[0]
        subElements = element[1]

        if subElements:
            subName = subElements[0]
            sub = ob.getSubObject(subName)

            if not sub or not hasattr(sub, "CenterOfMass"):
                return
            com = sub.CenterOfMass

        else:
            # Get Center of Mass of the object
            if not hasattr(ob, "Shape"):
                return

            shape = ob.Shape
            if not hasattr(shape, "CenterOfMass"):
                return

            com = shape.CenterOfMass

        obj.Result = com
        placement = Placement()
        placement.Base = com
        obj.Placement = placement

    def onChanged(self, obj, prop):
        """Do something when a property has changed"""

        if prop == "Element":
            self.execute(obj)



# ***************************************************************************
# *   Copyright (c) 2016 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM mesh group task panel for the document object"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"

## @package task_mesh_group
#  \ingroup FEM
#  \brief task panel for mesh group object

from PySide import QtCore

import FreeCAD
import FreeCADGui

from femguiutils import selection_widgets
from . import base_femtaskpanel


class _TaskPanel(base_femtaskpanel._BaseTaskPanel):
    """
    The TaskPanel for editing References property of MeshGroup objects
    """

    def __init__(self, obj):
        super().__init__(obj)

        # parameter widget
        self.parameterWidget = FreeCADGui.PySideUic.loadUi(
            FreeCAD.getHomePath() + "Mod/Fem/Resources/ui/MeshGroup.ui"
        )
        QtCore.QObject.connect(
            self.parameterWidget.rb_name,
            QtCore.SIGNAL("toggled(bool)"),
            self.choose_exportidentifier_name,
        )
        QtCore.QObject.connect(
            self.parameterWidget.rb_label,
            QtCore.SIGNAL("toggled(bool)"),
            self.choose_exportidentifier_label,
        )
        self.init_parameter_widget()

        # geometry selection widget
        # start with Solid in list!
        # only one shape type is allowed
        self.selectionWidget = selection_widgets.GeometryElementsSelection(
            obj.References, ["Solid", "Face", "Edge", "Vertex"], False, False
        )

        # form made from param and selection widget
        self.form = [self.parameterWidget, self.selectionWidget]

    def accept(self):
        self.obj.UseLabel = self.use_label
        self.obj.References = self.selectionWidget.references
        self.selectionWidget.finish_selection()
        return super().accept()

    def reject(self):
        self.selectionWidget.finish_selection()
        return super().reject()

    def init_parameter_widget(self):
        self.use_label = self.obj.UseLabel
        self.parameterWidget.rb_name.setChecked(not self.use_label)
        self.parameterWidget.rb_label.setChecked(self.use_label)

    def choose_exportidentifier_name(self, state):
        self.use_label = not state

    def choose_exportidentifier_label(self, state):
        self.use_label = state



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2014 Yorik van Havre <yorik@uncreated.net>              *
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides functions to create Layer objects."""
## @package make_layer
# \ingroup draftmake
# \brief Provides functions to create Layer objects.

## \addtogroup draftmake
# @{
import FreeCAD as App
from draftobjects.layer import Layer, LayerContainer
from draftutils import utils
from draftutils.messages import _err
from draftutils.translate import translate

if App.GuiUp:
    from draftviewproviders.view_layer import ViewProviderLayer, ViewProviderLayerContainer


def get_layer_container():
    """Return a group object to put layers in.

    Returns
    -------
    App::DocumentObjectGroupPython
        The existing group object named `'LayerContainer'`
        of type `LayerContainer`.
        If it doesn't exist it will create it with this default Name.
    """
    found, doc = utils.find_doc(App.activeDocument())
    if not found:
        _err(translate("draft", "No active document. Aborting."))
        return None

    for obj in doc.Objects:
        if obj.Name == "LayerContainer":
            return obj

    obj = doc.addObject("App::DocumentObjectGroupPython", "LayerContainer")
    obj.Label = translate("draft", "Layers")

    LayerContainer(obj)

    if App.GuiUp:
        ViewProviderLayerContainer(obj.ViewObject)

    return obj


def getLayerContainer():
    """Get the Layer container. DEPRECATED. Use 'get_layer_container'."""
    utils.use_instead("get_layer_container")

    return get_layer_container()


def make_layer(
    name=None,
    line_color=(0.0, 0.0, 0.0),  # does not match default DefaultShapeLineColor
    shape_color=(0.8, 0.8, 0.8),  # does not match default DefaultShapeColor
    line_width=2.0,
    draw_style="Solid",
    transparency=0,
):
    """Create a Layer object in the active document.

    If a layer container named `'LayerContainer'` does not exist, it is created.

    A layer controls the view properties of the objects inside the layer.
    All parameters except for `name` only apply if the graphical interface
    is up.

    All parameters that control view properties can be set to `None`. Their
    value, as set by the view provider (matching the current preferences), is
    then not changed.

    Parameters
    ----------
    name: str or `None`, optional
        It defaults to `None`.
        It is used to set the layer's `Label`. If it is `None` the `Label` is
        set to `'Layer'` or to its translation in the current language.

    line_color: tuple or `None`, optional
        It defaults to `(0.0, 0.0, 0.0)`.
        If it is given, it should be a tuple of three floating point values
        from 0.0 to 1.0.

    shape_color: tuple or `None`, optional
        It defaults to `(0.8, 0.8, 0.8)`.
        If it is given, it should be a tuple of three floating point values
        from 0.0 to 1.0.

    line_width: float or `None`, optional
        It defaults to 2.0.
        It determines the width of the edges of the objects contained in the layer.

    draw_style: str or `None`, optional
        It defaults to `'Solid'`.
        It determines the style of the edges of the objects contained in the layer.
        If it is given, it should be 'Solid', 'Dashed', 'Dotted' or 'Dashdot'.

    transparency: int or `None`, optional
        It defaults to 0.
        It should be an integer from 0 to 100.

    Return
    ------
    App::FeaturePython
        A scripted object of type `'Layer'`.
        This object does not have a `Shape` attribute.
        Modifying the view properties of this object will affect the objects
        inside of it.

    None
        If there is a problem it will return `None`.
    """
    _name = "make_layer"

    found, doc = utils.find_doc(App.activeDocument())
    if not found:
        _err(translate("draft", "No active document. Aborting."))
        return None

    if name is not None:
        try:
            utils.type_check([(name, str)], name=_name)
        except TypeError:
            _err(translate("draft", "Wrong input: it must be a string."))
            return None
    else:
        name = translate("draft", "Layer")

    if line_color is not None:
        try:
            utils.type_check([(line_color, tuple)], name=_name)
        except TypeError:
            _err(translate("draft", "Wrong input: must be a tuple of three floats 0.0 to 1.0."))
            return None

        if not all(isinstance(color, (int, float)) for color in line_color):
            _err(translate("draft", "Wrong input: must be a tuple of three floats 0.0 to 1.0."))
            return None

    if shape_color is not None:
        try:
            utils.type_check([(shape_color, tuple)], name=_name)
        except TypeError:
            _err(translate("draft", "Wrong input: must be a tuple of three floats 0.0 to 1.0."))
            return None

        if not all(isinstance(color, (int, float)) for color in shape_color):
            _err(translate("draft", "Wrong input: must be a tuple of three floats 0.0 to 1.0."))
            return None

    if line_width is not None:
        try:
            utils.type_check([(line_width, (int, float))], name=_name)
            line_width = float(abs(line_width))
        except TypeError:
            _err(translate("draft", "Wrong input: must be a number."))
            return None

    if draw_style is not None:
        try:
            utils.type_check([(draw_style, str)], name=_name)
        except TypeError:
            _err(
                translate(
                    "draft", "Wrong input: must be 'Solid', 'Dashed', 'Dotted', or 'Dashdot'."
                )
            )
            return None

        if draw_style not in ("Solid", "Dashed", "Dotted", "Dashdot"):
            _err(
                translate(
                    "draft", "Wrong input: must be 'Solid', 'Dashed', 'Dotted', or 'Dashdot'."
                )
            )
            return None

    if transparency is not None:
        try:
            utils.type_check([(transparency, (int, float))], name=_name)
            transparency = int(abs(transparency))
        except TypeError:
            _err(translate("draft", "Wrong input: must be a number between 0 and 100."))
            return None

    obj = doc.addObject("App::FeaturePython", "Layer")
    Layer(obj)

    obj.Label = name

    if App.GuiUp:
        vobj = obj.ViewObject
        ViewProviderLayer(vobj)
        if line_color is not None:
            vobj.LineColor = line_color
        if shape_color is not None:
            vobj.ShapeColor = shape_color
        if line_width is not None:
            vobj.LineWidth = line_width
        if draw_style is not None:
            vobj.DrawStyle = draw_style
        if transparency is not None:
            vobj.Transparency = transparency

    container = get_layer_container()
    container.addObject(obj)

    return obj


## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *   Copyright (c) 2021 FreeCAD Developers                                 *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

"""Provides functions and lists of commands to set up Draft menus and toolbars."""

## @package init_tools
# \ingroup draftutils
# \brief Provides lists of commands to set up toolbars of the workbench.

## \addtogroup draftutils
# @{
from PySide.QtCore import QT_TRANSLATE_NOOP

from draftutils.translate import translate

# Comment out commands that aren't ready to be used


def get_draft_drawing_commands():
    """Return the drawing commands list."""
    from draftguitools import gui_arcs
    from draftguitools import gui_beziers

    arc_group = gui_arcs.ArcGroup
    bez_group = gui_beziers.BezierGroup

    return [
        "Draft_Line",
        "Draft_Wire",
        "Draft_Fillet",
        (
            [QT_TRANSLATE_NOOP("Workbench", "Arc Tools")],
            list(arc_group.GetCommands(arc_group)),
        ),  # tuple len=2: submenu
        ("Draft_ArcTools",),  # tuple len=1: toolbar flyout
        "Draft_Circle",
        "Draft_Ellipse",
        "Draft_Rectangle",
        "Draft_Polygon",
        "Draft_BSpline",
        ([QT_TRANSLATE_NOOP("Workbench", "Bézier Tools")], list(bez_group.GetCommands(bez_group))),
        ("Draft_BezierTools",),
        "Draft_Point",
        "Draft_Facebinder",
        "Draft_ShapeString",
        "Draft_Hatch",
    ]


def get_draft_annotation_commands():
    """Return the annotation commands list."""
    return ["Draft_Text", "Draft_Dimension", "Draft_Label", "Draft_AnnotationStyleEditor"]


def get_draft_modification_commands():
    """Return the modification commands list."""
    from draftguitools import gui_arrays

    arr_group = gui_arrays.ArrayGroup

    return [
        "Draft_Move",
        "Draft_Rotate",
        "Draft_Scale",
        "Draft_Mirror",
        "Draft_Offset",
        "Draft_Trimex",
        "Draft_Stretch",
        "Separator",
        "Draft_Clone",
        (
            [QT_TRANSLATE_NOOP("Workbench", "Array Tools")],
            list(arr_group.GetCommands(arr_group)),
        ),  # tuple len=2: submenu
        ("Draft_ArrayTools",),  # tuple len=1: toolbar flyout
        "Separator",
        "Draft_Edit",
        "Draft_SubelementHighlight",
        "Separator",
        "Draft_Join",
        "Draft_Split",
        "Draft_Upgrade",
        "Draft_Downgrade",
        "Separator",
        "Draft_WireToBSpline",
        "Draft_Draft2Sketch",
        "Draft_Slope",
        "Draft_FlipDimension",
        "Separator",
        "Draft_Shape2DView",
    ]


def get_draft_utility_commands_menu():
    """Return the utility commands list for the menu."""
    return [
        "Draft_SetStyle",
        "Draft_ApplyStyle",
        "Separator",
        "Draft_Layer",
        "Draft_LayerManager",
        "Draft_AddNamedGroup",
        "Draft_SelectGroup",
        "Draft_ToggleConstructionMode",
        "Separator",
        "Draft_AddToLayer",
        "Draft_AddToGroup",
        "Draft_AddConstruction",
        "Separator",
        "Draft_ToggleDisplayMode",
        "Draft_ToggleGrid",
        "Draft_SelectPlane",
        "Draft_WorkingPlaneProxy",
        "Separator",
        "Draft_Heal",
        "Draft_ShowSnapBar",
    ]


def get_draft_utility_commands_toolbar():
    """Return the utility commands list for the toolbar."""
    return [
        "Draft_LayerManager",
        "Draft_AddNamedGroup",
        "Draft_SelectGroup",
        "Draft_AddToLayer",
        "Draft_AddToGroup",
        "Draft_AddConstruction",
        "Draft_ToggleDisplayMode",
        "Draft_WorkingPlaneProxy",
    ]


def get_draft_snap_commands():
    """Return the snapping commands list."""
    return [
        "Draft_Snap_Lock",
        "Draft_Snap_Endpoint",
        "Draft_Snap_Midpoint",
        "Draft_Snap_Center",
        "Draft_Snap_Angle",
        "Draft_Snap_Intersection",
        "Draft_Snap_Perpendicular",
        "Draft_Snap_Extension",
        "Draft_Snap_Parallel",
        "Draft_Snap_Special",
        "Draft_Snap_Near",
        "Draft_Snap_Ortho",
        "Draft_Snap_Grid",
        "Draft_Snap_WorkingPlane",
        "Draft_Snap_Dimensions",
        # "Separator", # Removed: if the Python generated BIM snap toolbar
        # is displayed in the Draft WB the separator appears
        # after the last button. Can be reinstated when the
        # BIM WB has a `normal` snap toolbar as well.
        "Draft_ToggleGrid",
    ]


def get_draft_context_commands():
    """Return the context menu commands list."""
    return [
        "Draft_SetStyle",
        "Draft_ApplyStyle",
        "Separator",
        "Draft_Layer",
        "Draft_LayerManager",
        "Draft_AddNamedGroup",
        "Draft_SelectGroup",
        "Draft_ToggleConstructionMode",
        "Separator",
        "Draft_AddToLayer",
        "Draft_AddToGroup",
        "Draft_AddConstruction",
        "Separator",
        "Draft_ToggleDisplayMode",
        "Draft_ToggleGrid",
        "Draft_SelectPlane",
        "Draft_WorkingPlaneProxy",
    ]


def init_toolbar(workbench, toolbar, cmd_list):
    """Initialize a toolbar.

    Parameters
    ----------
    workbench: Gui.Workbench
        The workbench. The commands from cmd_list must be available.

    toolbar: string
        The name of the toolbar.

    cmd_list: list of strings or list of strings and tuples
        See f.e. the return value of get_draft_drawing_commands.
    """
    for cmd in cmd_list:
        if isinstance(cmd, tuple):
            if len(cmd) == 1:
                workbench.appendToolbar(toolbar, [cmd[0]])
        else:
            workbench.appendToolbar(toolbar, [cmd])


def init_menu(workbench, menu_list, cmd_list):
    """Initialize a menu.

    Parameters
    ----------
    workbench: Gui.Workbench
        The workbench. The commands from cmd_list must be available.

    menu_list: list of strings
        The main and optional submenu(s). The commands, and additional
        submenus (if any), are added to the last (sub)menu in the list.

    cmd_list: list of strings or list of strings and tuples
        See f.e. the return value of get_draft_drawing_commands.
    """
    for cmd in cmd_list:
        if isinstance(cmd, tuple):
            if len(cmd) == 2:
                workbench.appendMenu(menu_list + cmd[0], cmd[1])
        else:
            workbench.appendMenu(menu_list, [cmd])


## @}
